# Copyright (c) 2021-2023  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import math
from typing import Optional

# Third party imports
from skimage import filters
from skimage.transform import downscale_local_mean
import numpy as np
import numpy.typing as npt

# Local application imports
# from navigate.model.analysis.camera import compute_signal_to_noise


def has_tissue(
    image_data: npt.ArrayLike,
    x: int,
    y: int,
    width: int,
    offset: Optional[npt.ArrayLike] = None,
    variance: Optional[npt.ArrayLike] = None,
) -> bool:
    """
    Determine if an image contains tissue.

    Parameters
    ----------
    image_data : npt.ArrayLike
        Image
    x : int
        Starting x-position of subimage. Must be smaller than image_data.shape[0].
    y : int
        Starting y-position of subimage. Must be smaller than image_data.shape[1].
    width : int
        Width of subimage. Must be smaller than min(image_data.shape[:1])
    offset : npt.ArrayLike
        Camera pixel offset map. Same size as image_data.
    variance : npt.ArrayLike
        Camera pixel variance map. Same size as image_data.

    Returns
    -------
    bool
        Is tissue present?
    """
    # TODO: threshold value, other is_tissue() method
    xsl = slice(x * width, (x + 1) * width)
    ysl = slice(y * width, (y + 1) * width)

    # Decide tissue is present by hard SNR threshold
    # if offset is not None and variance is not None:
    #     threshold_value = 2  # snr threshold
    #     snr = compute_signal_to_noise(
    #         image_data[xsl, ysl], offset[xsl, ysl], variance[xsl, ysl]
    #     )
    #     return (
    #         np.sum(snr > threshold_value) / np.prod(snr.shape) > 0.0625
    #     )  # at least 6.25% of pixels have to be above this value

    # Decide tissue is present by hard pixel count threshold
    # threshold_value = 1000
    return np.any(image_data[xsl, ysl] > np.mean(image_data))


def find_tissue_boundary_2d(
    image_data: npt.ArrayLike, mag_ratio: Optional[float] = 1.0
) -> list:
    """
    Find all pixels containing tissue, based on an Otsu threshold. Optionally,
    return pixels in image space resampled by mag_ratio.

    Parameters
    ----------
    image_data : npt.ArrayLike
        Image
    mag_ratio : float
        Ratio between pixel sizes of current over target tiles.

    Returns
    -------
    boundary : list
        List of boundaries of tissue by row of downsampled image.
    """

    # Threshold
    thresh_img = image_data > filters.threshold_otsu(image_data)

    if mag_ratio > 1:
        ds_img = downscale_local_mean(thresh_img, (mag_ratio, mag_ratio))
    else:
        ds_img = image_data
        mag_ratio = 1

    idx_x, idx_y = np.where(ds_img)

    # Assume square image
    m = math.ceil(image_data.shape[0] / mag_ratio)
    # n = math.ceil(image_data.shape[1] / mag_ratio)
    boundary = [None] * m
    for x, y in zip(idx_x, idx_y):
        if boundary[x] is None:
            boundary[x] = [y, y]
        else:
            boundary[x][1] = y
    return boundary


def binary_detect(
    img_data: npt.ArrayLike,
    boundary: list,
    width: int = 1,
    offset: Optional[npt.ArrayLike] = None,
    variance: Optional[npt.ArrayLike] = None,
):
    """
    Perform a binary search for tissue on an image. Return locations of pixels
    containing tissue.

    Parameters
    ----------
    img_data : npt.ArrayLike
        Image
    boundary : dict
        List of xy pixel positions indicating presence of tissue as values.
    width : int
        Width of subimage. Must be smaller than min(image_data.shape[:1])
    offset : npt.ArrayLike
        Camera pixel offset map. Same size as image_data.
    variance : npt.ArrayLike
        Camera pixel variance map. Same size as image_data.

    Returns
    -------
    boundary : list
        List of boundaries of tissue by row of downsampled image.
    """
    m, n = img_data.shape
    m = int(m / width)
    n = int(n / width)

    def binary_search_func_left(row, left, right):
        """Binary search function.

        Parameters
        ----------
        row : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Leftmost column index of subimage.
        """
        while left < right:
            mid = (left + right) // 2
            if has_tissue(img_data, row, mid, width, offset, variance):
                right = mid
            else:
                left = mid + 1
        return right

    def binary_search_func_right(row, left, right):
        """Binary search function.

        Parameters
        ----------
        row : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Rightmost column index of subimage.
        """

        while left < right:
            mid = (left + right) // 2
            if has_tissue(img_data, row, mid, width, offset, variance):
                left = mid + 1
            else:
                right = mid
        return right - 1

    def find_tissue_range(row, left, right):
        """Find range of tissue in image.

        Parameters
        ----------
        row : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Leftmost column index of subimage.
        int
            Middle column index of subimage.
        int
            Rightmost column index of subimage.
        """

        temp = [(left, right)]
        while temp:
            temp2 = []
            for ll, r in temp:
                mid = (ll + r) // 2
                if has_tissue(img_data, row, mid, width, offset, variance):
                    return ll, mid, r
                if mid > ll + 1:
                    temp2.append((ll, mid))
                if r > mid + 1:
                    temp2.append((mid, r))
            temp = temp2
        return -1, -1, -1

    def detect_row_boundary(row_id, left, right):
        """Detect row boundary.

        Parameters
        ----------
        row_id : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Leftmost column index of subimage.
        int
            Rightmost column index of subimage.
        """
        is_tissue_left = has_tissue(img_data, row_id, left, width, offset, variance)
        is_tissue_right = has_tissue(img_data, row_id, right, width, offset, variance)

        if is_tissue_left and is_tissue_right:
            left_l, left_r = 0, left
            right_l, right_r = right, n
        elif is_tissue_left:
            left_l, left_r = 0, left
            right_l, right_r = left, right
        elif is_tissue_right:
            left_l, left_r = left, right
            right_l, right_r = right, n
        else:
            ll, mid, r = find_tissue_range(row_id, left, right)
            left_l, left_r = ll, mid
            right_l, right_r = mid, r

        if left_l == -1:
            return None, None
        return binary_search_func_left(
            row_id, left_l, left_r
        ), binary_search_func_right(row_id, right_l, right_r)

    def expand_row(row_id, limits, direction, boundary):
        """Expand row.

        Parameters
        ----------
        row_id : int
            Row index of image.
        limits : int
            Limits of row.
        direction : int
            Direction of row.
        boundary : list
            List of boundaries of tissue by row of downsampled image.

        Returns
        -------
        new_boundary : list
            List of boundaries of tissue by row of downsampled image.
        """

        for i in range(row_id, limits, direction):
            left, right = boundary[i][0], boundary[i][1]
            left = left - 1 if left > 0 else 0
            right = right + 1 if (right + 1) < (n - 1) else (n - 1)
            ll, r = detect_row_boundary(i + direction, left, right)
            if ll is None:
                boundary[i + direction] = None
                break
            boundary[i + direction] = [ll, r]

    new_boundary = boundary[:]
    top, bottom = None, None
    expand_top, expand_bottom = True, True

    for i, row in enumerate(new_boundary):
        if row is None:
            continue
        if expand_bottom is False:
            new_boundary[i] = None
            continue
        if len(row) < 2:
            row.append(row[0])
        left, right = detect_row_boundary(i, row[0], row[1])
        if left is None:
            new_boundary[i] = None
            if top is None:
                expand_top = False
            if bottom is not None:
                expand_bottom = False
        else:
            new_boundary[i] = [left, right]

        if top is None:
            top = i
        bottom = i

    # detect top/bottom if necessary
    if expand_top:
        expand_row(top, 0, -1, new_boundary)

    if expand_bottom:
        expand_row(bottom, n - 1, 1, new_boundary)

    return new_boundary


def map_boundary(boundary, direction=True):
    """Map boundary to a path.

    Parameters
    ----------
    boundary : list
        List of boundaries of tissue by row of downsampled image.
    direction : bool
        Direction of path.

    Returns
    -------
    path : list
        List of boundaries of tissue by row of downsampled image.
    """
    if direction:
        start, end, step = 0, len(boundary), 1
        offset = -1
    else:
        start, end, step = len(boundary) - 1, -1, -1
        offset = 1

    def dp_shortest_path(start, end, step, offset=-1):
        """Dynamic programming shortest path.

        Parameters
        ----------
        start : int
            Starting index.
        end : int
            Ending index.

        Returns
        -------
        path : list
            List of boundaries of tissue by row of downsampled image.
        """
        dp_path = []
        dp_cost = [0, 0]
        visited = False
        for i in range(start, end, step):
            if boundary[i] is None:
                if visited:
                    break
                continue
            w = boundary[i][1] - boundary[i][0] + 1
            if not visited:
                visited = True
                dp_cost[0], dp_cost[1] = w, w
                dp_path.append([i, -1, -1])
            else:
                dp_path.append([i, 0, 0])
                for j in range(2):
                    ll = abs(boundary[i + offset][0] - boundary[i][j])
                    r = abs(boundary[i + offset][1] - boundary[i][j])
                    if ll < r:
                        dp_cost[1 - j] += w + ll
                        dp_path[-1][2 - j] = 0
                    else:
                        dp_cost[1 - j] += w + r
                        dp_path[-1][2 - j] = 1

        # reverse path
        if dp_cost[0] < dp_cost[1]:
            idx = 0
        else:
            idx = 1
        path = []
        for item in reversed(dp_path):
            x = item[0]
            if idx == 0:
                path += map(lambda y: (x, y), range(boundary[x][0], boundary[x][1] + 1))
            else:
                path += map(
                    lambda y: (x, y), range(boundary[x][1], boundary[x][0] - 1, -1)
                )
            idx = item[idx + 1]
        path.reverse()
        return path

    result = dp_shortest_path(start, end, step, offset)
    return result



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from queue import Queue
from navigate.model.analysis.boundary_detect import (
    find_tissue_boundary_2d,
    binary_detect,
    map_boundary,
)
import numpy as np


def draw_box(img, xl, yl, xu, yu, fill=65535):
    """Draw a box on an image

    Parameters
    ----------
    img : np.ndarray
        Image to draw on
    xl : int
        Lower x coordinate
    yl : int
        Lower y coordinate
    xu : int
        Upper x coordinate
    yu : int
        Upper y coordinate
    fill : int
        Value to fill box with

    Returns
    -------
    np.ndarray
        Image with box drawn
    """
    img[xl:xu, yl] = fill
    img[xl:xu, yu] = fill
    img[xl, yl:yu] = fill
    img[xu, yl:yu] = fill
    return img


class VolumeSearch:
    """VolumeSearch.

    Find the outer boundary of a tissue, moving the stage through z. Assumes
    there is no tissue out-of-frame in the x,y-directions.

    Construct a list of stage positions that will tile the entire tissue
    boundary in x,y,z at target_resolution, target_zoom. Current resolution
    and zoom is self.model.active_microscope_name and
    self.model.configuration["experiment"]["MicroscopeState"]["zoom"].

    """

    def __init__(
        self,
        model,
        target_resolution="Nanoscale",
        target_zoom="N/A",
        flipx=False,
        flipy=False,
        overlap=0.1,
        debug=False,
    ):
        """Initialize VolumeSearch

        Parameters
        ----------
        model : navigate.model.model.Model
            Navigate Model
        target_resolution : str
            Name of microscope to use for tiled imaging of tissue
        target_zoom : str
            Resolution of microscope (target_resolution) to use for tiled imaging
            of tissue
        flipx : bool
            Flip the direction in which new tiles are added.
        flipy : bool
            Flip the direction in which new tiles are added.
        overlap : float
            Value between 0 and 1 indicating percent overlap of tiles.
        debug : bool
            If True, save debug images to disk.
        """

        #: navigate.model.model.Model: Navigate Model
        self.model = model

        #: str: Name of microscope to use for tiled imaging of tissue
        self.target_resolution = target_resolution

        #: str: Resolution of microscope (target_resolution) to use for tiled imaging
        self.target_zoom = target_zoom

        #: int: Number of z-steps to take
        self.z_steps = 0

        #: float: Size of z-step
        self.z_step_size = 0

        #: float: Current z-position
        self.z_pos = 0

        #: float: Size of focus step
        self.f_step_size = 0

        #: float: Current focus position
        self.f_pos = 0

        #: int: Current z-index
        self.f_offset = 0

        #: int: Current z-index
        self.curr_z_index = 0

        # By default an increase in x/y stage position corresponds
        # to a sample moving down/right into the field of view
        #: int: 1 if flipx is False, -1 if flipx is True
        self.sinx = 1 if flipx else -1

        #: int: 1 if flipy is False, -1 if flipy is True
        self.siny = 1 if flipy else -1

        #: float: Percent overlap of tiles
        self.overlap = max(0, min(overlap, 0.999))

        #: Queue: Queue for communicating whether tissue was found
        self.has_tissue_queue = Queue()

        #: int: 1 if moving up, -1 if moving down
        self.direction = 1  # up: 1; down: -1

        #: bool: True if tissue was found
        self.has_tissue = False

        #: dict: Boundary of tissue
        self.first_boundary = None

        #: dict: Boundary of tissue
        self.pre_boundary = None

        #: bool: True if the search is complete
        self.end_flag = False

        #: dict: Feature configuration
        self.config_table = {
            "signal": {
                "init": self.pre_signal_func,
                "main": self.signal_func,
                "main-response": self.signal_response_func,
                "end": self.signal_end,
                "cleanup": self.cleanup,
            },
            "data": {
                "init": self.init_data_func,
                "main": self.data_func,
                "end": self.end_data_func,
                "cleanup": self.cleanup,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }

        #: bool: True if debug mode is enabled
        self.debug = debug

        #: np.ndarray: Debug image
        self.volumes_selected = None

    def pre_signal_func(self):
        """Initialize signal function"""
        self.model.active_microscope.current_channel = 0
        self.model.active_microscope.prepare_next_channel()

        self.z_pos = float(
            self.model.configuration["experiment"]["StageParameters"]["z"]
        )
        self.f_pos = float(
            self.model.configuration["experiment"]["StageParameters"]["f"]
        )

        self.z_steps = float(
            self.model.configuration["experiment"]["MicroscopeState"]["number_z_steps"]
        )
        self.z_step_size = float(
            self.model.configuration["experiment"]["MicroscopeState"]["step_size"]
        ) * (1 - self.overlap)

        f_start = float(
            self.model.configuration["experiment"]["MicroscopeState"]["start_focus"]
        )
        f_end = float(
            self.model.configuration["experiment"]["MicroscopeState"]["end_focus"]
        )
        self.f_step_size = (f_end - f_start) / self.z_steps * (1 - self.overlap)

        self.curr_z_index = int(self.z_steps / 2)

        self.direction = 1  # up

        if self.debug:
            self.volumes_selected = np.zeros(
                (
                    int(self.z_steps),
                    self.model.img_height,
                    self.model.img_width,
                ),
                dtype="uint16",
            )

    def signal_func(self):
        """Signal function.

        Move the stage to the next z-position.

        Returns
        -------
        bool
            True if the signal function should be called again.
        """
        self.model.logger.debug(f"acquiring at z:{self.curr_z_index}")
        z = self.z_pos + self.curr_z_index * self.z_step_size
        f = self.f_pos + self.curr_z_index * self.f_step_size
        self.model.move_stage({"z_abs": z, "f_abs": f})
        return True

    def signal_response_func(self, *args):
        """Signal response function.

        Check if tissue was found. If not, move the stage to the next z-position.

        Returns
        -------
        bool
            True if the signal function should be called again.
        """
        has_tissue = self.has_tissue_queue.get()

        if not has_tissue or self.curr_z_index == self.z_steps - 1:
            self.curr_z_index = int(self.z_steps / 2)
            self.direction = -1

        self.curr_z_index += self.direction
        return True

    def signal_end(self):
        """Signal end function.

        Check if the search is complete.

        Returns
        -------
        bool
            True if the search is complete.
        """
        return self.end_flag

    def init_data_func(self):
        """Initialize data function"""
        # Establish current and target pixel sizes
        microscope_name = self.model.active_microscope_name
        curr_zoom = self.model.configuration["experiment"]["MicroscopeState"]["zoom"]
        curr_pixel_size = float(
            self.model.configuration["configuration"]["microscopes"][microscope_name][
                "zoom"
            ]["pixel_size"][curr_zoom]
        )
        target_pixel_size = float(
            self.model.configuration["configuration"]["microscopes"][
                self.target_resolution
            ]["zoom"]["pixel_size"][self.target_zoom]
        )

        # consider the image as a square
        img_width = self.model.configuration["experiment"]["CameraParameters"][
            "x_pixels"
        ]

        # The target image size in pixels
        self.mag_ratio = int(curr_pixel_size / target_pixel_size)
        self.target_grid_pixels = int(img_width // self.mag_ratio)
        # The target image size in microns
        self.target_grid_width = img_width * target_pixel_size * (1 - self.overlap)

        # For each axis, establish the offset between this image and the target
        # image as the difference in the physical offsets of the two microscopes
        axes = ["x", "y", "z", "theta", "f"]
        self.offset = [0, 0, 0, 0, 0]
        for i, axis in enumerate(axes):
            t = axis + "_offset"
            self.offset[i] = float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"][t]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"][t]
            )

        # Set this to the upper left corner of the image
        self.offset[0] += (
            self.model.configuration["experiment"]["StageParameters"]["x"]
            - self.sinx * (img_width - self.target_grid_pixels) // 2 * curr_pixel_size
        )
        self.offset[1] += (
            self.model.configuration["experiment"]["StageParameters"]["y"]
            - self.siny * (img_width - self.target_grid_pixels) // 2 * curr_pixel_size
        )
        self.offset[2] += self.z_pos
        self.offset[3] += self.model.configuration["experiment"]["StageParameters"][
            "theta"
        ]

        offsets = self.model.active_microscope.zoom.stage_offsets
        focus_offset = 0
        if offsets is not None:
            solvent = self.model.configuration["experiment"]["Saving"]["solvent"]
            try:
                focus_offset = offsets[solvent]["f"][curr_zoom][self.target_zoom]
            except Exception:
                focus_offset = 0
        self.offset[4] += self.f_pos + focus_offset

        self.first_boundary = None
        self.pre_boundary = None
        self.boundary = {}

    def data_func(self, frame_ids):
        """Data function.

        Find the tissue boundary in the current frame.

        Parameters
        ----------
        frame_ids : list
            List of frame ids to process.

        Returns
        -------
        bool
            True if the data function should be called again.
        """
        for idx in frame_ids:
            img_data = self.model.data_buffer[idx]
            # TODO: make sure set the right threshold_value in
            # model\analysis\boundary_detect.py when use if/else
            # boundary = find_tissue_boundary_2d(img_data, self.target_grid_pixels)

            if self.debug:
                self.volumes_selected[self.curr_z_index] = img_data

            if self.pre_boundary is None:
                boundary = find_tissue_boundary_2d(img_data, self.target_grid_pixels)
            else:
                off, var = self.model.get_offset_variance_maps()
                boundary = binary_detect(
                    img_data, self.pre_boundary, self.target_grid_pixels, off, var
                )

            self.has_tissue = any(boundary)
            self.boundary[self.curr_z_index] = boundary

            if (
                self.has_tissue
                and self.curr_z_index > 0
                and self.curr_z_index < self.z_steps - 1
            ):
                self.pre_boundary = boundary
            elif self.direction == 1:
                self.pre_boundary = self.first_boundary
            else:
                self.end_flag = True

            self.model.logger.debug(
                f"has tissue? {self.curr_z_index} - {self.has_tissue}"
            )
            self.has_tissue_queue.put(self.has_tissue)

    def end_data_func(self):
        """End data function.

        Map the tissue boundary to stage positions.

        Returns
        -------
        bool
            True if the search is complete.
        """
        if self.end_flag:
            direction = True
            positions = []
            for z_index in sorted(self.boundary.keys()):
                path = map_boundary(self.boundary[z_index], direction)
                direction = not direction
                positions += map(
                    lambda item: (
                        self.sinx * item[0] * self.target_grid_width + self.offset[0],
                        self.siny * item[1] * self.target_grid_width + self.offset[1],
                        z_index * self.z_step_size + self.offset[2],
                        self.offset[3],
                        z_index * self.f_step_size + self.offset[4],
                    ),
                    path,
                )
                if self.debug:
                    for item in path:
                        self.volumes_selected[z_index] = draw_box(
                            self.volumes_selected[z_index],
                            int(item[0] * self.target_grid_pixels * (1 - self.overlap)),
                            int(item[1] * self.target_grid_pixels * (1 - self.overlap)),
                            int(
                                item[0] * self.target_grid_pixels * (1 - self.overlap)
                                + self.target_grid_pixels
                            )
                            - 1,
                            int(
                                item[1] * self.target_grid_pixels * (1 - self.overlap)
                                + self.target_grid_pixels
                            )
                            - 1,
                        )
            self.model.event_queue.put(("multiposition", positions))
            if self.debug:
                import tifffile

                tifffile.imsave(
                    "volume_search_2d_debug_result.tif", self.volumes_selected
                )
        return self.end_flag

    def cleanup(self):
        """Cleanup function"""
        self.has_tissue_queue.put(False)



import logging

import serial
import re

# from time import time

from navigate.model.devices.lasers.laser_base import LaserBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class LuxxLaser(LaserBase):
    def __init__(self, comport):
        """
        # Luxx Laser Class
        # LUXX488, 200 mW, is COM19
        # LUXX642, 140 mW, is COM17
        # LIGHTHUB ULTRA, is COM20
        # Open port (*auto* stands for **/dev/ttyUSB0** in Linux or **COM17**
        # in Windows, because it is what I use); then get device model;
        # finally, get maximum output power and store it in **pmax** variable.
        """
        self.comport = comport
        self.baudrate = 500000
        self.timeout = 0.3

        try:
            # Open serial port
            self.laser = serial.Serial(
                self.comport, self.baudrate, timeout=self.timeout
            )

            logger.debug(f"The laser is connected via {self.comport}")

            # Confirm the Laser Wavelength
            # Must remove non-standard ASCII codes that cause errors in the utf-8 codec
            wavelength = self.ask("GSI")
            wavelength = wavelength.replace(b"\xa7200", str(" ").encode())
            wavelength = wavelength.replace(b"\xa7140", str(" ").encode())
            wavelength = float(wavelength.decode())
            self.wavelength = wavelength
            logger.debug(f"The Wavelength is: {wavelength}")

            # Confirm the Laser Serial Number
            serial_number = self.ask("GSN")
            serial_number = serial_number.decode()
            self.serial = serial_number
            logger.debug(f"The laser Serial Number: {serial_number}")

            # Confirm the Laser Working Hours
            laser_working_hours = self.ask("GWH")
            laser_working_hours = laser_working_hours.decode()
            self.hours = laser_working_hours
            logger.debug(f"Laser working hours: {laser_working_hours}")

            # Confirm the Laser Maximum Power
            power_max = float(self.ask("GMP"))
            self.pmax = power_max
            logger.debug(f"Laser max power: {power_max}")

        except serial.SerialException:
            raise OSError(
                'Port "%s" is unavailable.\n' % self.comport
                + "Maybe the laser is not connected, the wrong"
                + " port is specified or the port is already opened"
            )

    def __del__(self):
        """
        # Close the port before exit.
        """
        try:
            # Turn off the laser emission
            # self.write("LOf")

            # Close the port
            self.laser.close()
            logger.debug("Port closed")

        except serial.SerialException:
            print("Could not close the port")

    def close(self):
        """
        # Close the port before exit.
        """
        try:
            self.laser.close()

        except serial.SerialException:
            print("could not close the port")

    def write(self, command):
        """
        # Send *command* to device. Preceed it with "?" und end with CR.
        """
        command = str("?").encode() + str(command).encode() + str("\r").encode()
        self.laser.write(command)

    def read(self):
        """
        # Read all information from the port and return it as string.
        """
        answer = self.laser.readall()
        answer = answer.replace(str("\r").encode(), str("\n").encode())
        answer = answer.replace(str("\xa7").encode(), str(" | ").encode())
        return answer

    def ask(self, command):
        """
        # Write, then read.
        # Uses regular expressions to clean up response from the device
        # !UK = Unknown command
        """
        self.write(command)
        response = self.read()
        search_string = str("!UK\n").encode()
        if re.findall(search_string, response):
            print("Command '%s' is unknown for this device" % command)
        else:
            search_string = "!%s(.+)\n" % command[:3]
            search_string = search_string.encode()
            print(f"The search string is: {search_string}")
            print(f"The response string is: {response}")

            if len(response) > 0:
                response = re.findall(search_string, response)[-1]

            if response[0] == "x":
                print("Laser responded with error to command '%s'" % command)

            return response

    def smart_ask(self, command):
        """
        #  TODO: Not working.  Depends on broken print_hex function.
        Several commands return information coded in ASCII HEX numbers.
        The relevant are bits in the registers. For convenient
        representation, we will print(these bytes in tables.
        This is relevant for the following commands:
            * GOM
            * GFB
            * GLF
            * GLP
        For all other commands the behavior is identical to **ask** function
        """
        if command in ["GOM", "GFB", "GLF", "GLP"]:
            return print_hex(self.ask(command))
        else:
            return self.ask(command)

    def start(self):
        """
        # Start the emission (takes about 3 seconds)
        """
        self.write("LOn")

    def stop(self):
        """
        # Stop the emission immediately
        """
        self.write("LOf")

    def set_power(self, power):
        """
        Set the desired power in mW
        TODO: Intermittent failure due to list index out of range in self.ask (line 115)
        """
        # Calculate the corresponding HEX code and transmit it
        if power > self.pmax:
            print("Laser provides %imW only. The maximum power is set", self.pmax)
            self.write("SLPFFF")
        else:
            code = hex(int(4095 * power / self.pmax))[2:].upper().zfill(3)
            stopwatch(self.ask, "SLP%s" % code)

    def get_power(self):
        """
        # Get the current power value in mW
        """
        code = stopwatch(self.ask, "GLP")
        return int(code, 16) * self.pmax / 4095.0

    def set_mode(self, mode):
        """
        # The device is able to work in the following modes:
        #  * Standby
        #      Laser is ready, but no emission is produced. However, if we
        #      it is turned on (e.g. with **start** function), then change to
        #      other mode will result in immediate emission, i.e. without
        #      3 seconds delay.
        #  * CW-ACC
        #      constant wave, automatic current control
        #  * CW-APC
        #      constant wave, automatic power control
        #  * Analog
        #      the output power is dependent on the analog input; however,
        #      it cannot exceed the specified with **set_power** value.
        """
        if mode == "Standby" or mode == 0:
            mode = 0
        elif mode == "CW-ACC" or mode == 1:
            mode = 1
        elif mode == "CW-APC" or mode == 2:
            mode = 2
        elif mode == "Analog" or mode == 3:
            mode = 3
        else:
            print(
                "**mode** must be one of 'Standby', 'CW-ACC', "
                + "'CW-APC', 'Analog' or number 0-3. Nothing changed."
            )
            return
        stopwatch(self.ask, "ROM%i" % mode)

    def get_mode(self):
        """
        # Get the current mode.
        """
        mode = int(stopwatch(self.ask, "ROM"))
        if mode == 0:
            return "Standby"
        elif mode == 1:
            return "CW-ACC"
        elif mode == 2:
            return "CW-APC"
        elif mode == 3:
            return "Analog"
        else:
            return mode

    def set_autostart(self, state):
        """
        # Decide if light is emitted on powerup.
        """
        if state:
            self.ask("SAS1")
        else:
            self.ask("SAS0")

    def get_autostart(self):
        """
        # Check if light is emitted on powerup.
        """
        if self.ask("GAS") == "1":
            return True
        else:
            return False

    def get_parameters(self):
        """
        # print a table showing laser status.
        # Bit description:
        # 15  Auto PowerUP if ONE
        # 14  Autostart (emission at powerup) if ONE
        # 13  Adhoc USB - Laser sends info messages from time to time if ONE
        # 8   Power control (APC) mode if ONE; current control (ACC) if ZERO
        # 7   External analog input enabled if ONE
        # 4   Mod level: ONE - active; ZERO - not active
        # 3   Bias level: ONE - active; ZERO - not active
        # Bits 0, 1, 2, 5, 6, 9, 10, 11, 12 are reserved
        """
        self.smart_ask("GOM")

    def get_errors(self):
        """
        # print contents of failure byte
        # Bit description:
        # 15  Diode power exceeded maximum value
        # 14  An internal error occured
        # 12  The temperature at the diode exceeded the valid temperature range
        # 11  The ambient temperature exceeded the minimum or maximum value
        # 10  The current through the diode exceeded the maximum allowed value
        # 9   The interlock loop is not closed. Please close the interlockt loop
        # 8   Overvoltage or Undervoltage lockout occured. Bring supply voltage to
        #     a valid range
        # 4   If CDRH-Bit is set and no CDRH-Kit is connected or
        #     CDRH-Bit is not set but a CDRH-Kit is connected
        #     CDRH-Kit is a box with a key, a LED and an interlock
        # 0   Soft interlock: If an interlock error occurs, this bit is set.
        #     It can only be reset by resetting the whole system, even if the
        #     interlock error is not present anymore.
        # Bits 1, 2, 3, 5, 6, 7, 13 are reserved
        """
        self.smart_ask("GFB")

    def initialize_laser(self):
        """
        # Initialize lasers.
        # Sets the laser to the maximum power, and sets the mode to CW-APC.
        """
        self.set_power(self.pmax)
        self.set_mode("CW-APC")
        self.start()
        logger.debug(
            f"{self.wavelength}nm Laser Initialized - Max Power: {self.pmax}mW"
        )


def stopwatch(func, *func_args, **func_kwargs):
    """
    # Call **func** and print elapsed time
    """
    # start_time = time()
    result = func(*func_args, **func_kwargs)
    # print("Time elapsed: %5.2f ms" % ((time() - start_time)*1000.0))
    return result


def print_hex(hex_code):
    """
    TODO: Not working.  Concatenation of list and range?
    print a nice table that represents *hex_code* (ASCII HEX string)
    in a binary code.

    Returns
    -------
    corresponding decimal numbers in array
    """
    # If the number is odd, pad it with leading zero
    length = len(hex_code)
    byte_number = (length / 2) + (length % 2)
    if length % 2:
        hex_code = "0" + hex_code

    # Split it into 8-bit numbers coded in ASCII HEX
    hex_numbers = []
    for i in range(int(byte_number)):
        hex_numbers.append(hex_code[i * 2 : i * 2 + 2])

    # Convert each of them into decimal
    decimals = []
    for hex_number in hex_numbers:
        decimals.append(int(hex_number, 16))

    # print a table
    table = """BYTE ##  :  '?'
    | ## | ## | ## | ## | ## | ## | ## | ## |
    |----|----|----|----|----|----|----|----|
    |  ? |  ? |  ? |  ? |  ? |  ? |  ? |  ? |
    """
    table = table.replace("##", "%2i").replace("?", "%s")

    print("\nRepresentation of ASCII HEX '%s'" % hex_code)
    for i, number in enumerate(decimals):
        byte = len(decimals) - i - 1

        # data_range is type Range
        data_range = range(byte * 8, byte * 8 + 8)[::-1]

        # data_list is type list
        data_list = list(bin(number)[2:].zfill(8))
        print("data list:", data_list)

        content = data_range + data_list
        print(table % tuple([byte, hex_numbers[i]] + content))
    return decimals



from pyvcam import pvc
from pyvcam.camera import Camera
from matplotlib import pyplot as plt

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()

    frame = cam.get_frame(exp_time=20).reshape(cam.sensor_size[::-1])
    plt.imshow(frame, cmap="gray")
    plt.show()
    cam.close()
    pvc.uninit_pvcam()

if __name__=="__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import numpy as np
import pandas as pd
from math import ceil


def sign(x):
    """Return the sign of x.

    Parameters
    ----------
    x : float
        The number to be signed.

    Returns
    -------
    int
        The sign of x.
    """
    # (1 if x > 0 else 0)
    return -1 if x < 0 else 1


def compute_tiles_from_bounding_box(
    x_start,
    x_tiles,
    x_length,
    x_overlap,
    y_start,
    y_tiles,
    y_length,
    y_overlap,
    z_start,
    z_tiles,
    z_length,
    z_overlap,
    theta_start,
    theta_tiles,
    theta_length,
    theta_overlap,
    f_start,
    f_tiles,
    f_length,
    f_overlap,
    f_track_with_z=False,
):
    """Create a grid of ROIs to image based on start position, number of tiles, and
    signed FOV length in each dimension.

    Parameters
    ----------
    x_start : float
        Starting position along x-dimension.
    x_tiles : int
        Number of tiles to take along x-dimension.
    x_length : float
        Signed length of the FOV along x-dimension.
    x_overlap : float
        Fractional overlap of ROIs along x-dimension.
    y_start : float
        Starting position along y-dimension.
    y_tiles : int
        Number of tiles to take along y-dimension.
    y_length : float
        Signed length of the FOV along y-dimension.
    y_overlap : float
        Fractional overlap of ROIs along y-dimension.
    z_start : float
        Starting position along z-dimension.
    z_tiles : int
        Number of tiles to take along z-dimension.
    z_length : float
        Signed length of the FOV along z-dimension.
    z_overlap : float
        Fractional overlap of ROIs along z-dimension.
    theta_start : float
        Starting position along rotation dimension.
    theta_tiles : int
        Number of tiles to take along rotation dimension.
    theta_length : float
        Signed length of the FOV along rotation dimension.
    theta_overlap : float
        Fractional overlap of ROIs along rotation dimension.
    f_start : float
        Starting position along focus dimension.
    f_tiles : int
        Number of tiles to take along focus dimension.
    f_length : float
        Signed length of the FOV along focus dimension.
    f_overlap : float
        Fractional overlap of ROIs along focus dimension.
    f_track_with_z : bool
        Make focus track with z/assume focus is z-dependent.

    Returns
    -------
    np.array
        (n_positions x (x, y, z, theta, f)) array of positions, gridding out the space
    """

    # Error checking to prevent empty list when tiles are zero
    x_tiles = 1 if x_tiles <= 0 else x_tiles
    y_tiles = 1 if y_tiles <= 0 else y_tiles
    z_tiles = 1 if z_tiles <= 0 else z_tiles
    theta_tiles = 1 if theta_tiles <= 0 else theta_tiles
    f_tiles = 1 if f_tiles <= 0 else f_tiles

    # Calculate the step between the edge of each frame
    x_step = x_length * (1 - x_overlap)
    y_step = y_length * (1 - y_overlap)
    z_step = z_length * (1 - z_overlap)
    theta_step = theta_length * (1 - theta_overlap)
    f_step = f_length * (1 - f_overlap)

    # grid out each dimension starting from (x_start, y_start, z_start) in steps
    def dim_vector(start, n_tiles, step):
        return start + np.arange(0, n_tiles, 1) * step

    xs = dim_vector(x_start, x_tiles, x_step)  # x-coordinate is centered on FOV
    ys = dim_vector(y_start, y_tiles, y_step)  # y-coordinate is centered on FOV
    zs = dim_vector(
        z_start, z_tiles, z_step
    )  # z-coordinate is centered on local z-stack origin
    thetas = dim_vector(
        theta_start, theta_tiles, theta_step
    )  # we assume theta FOVs have no thickness
    fs = dim_vector(f_start, f_tiles, f_step)

    if f_track_with_z:
        # grid out the 4D space...
        x, y, z, t = np.meshgrid(xs, ys, zs, thetas)

        # we need to make f vary the same as z, for now, since focus changes with z
        lz = len(z.ravel())
        f = np.repeat(fs, np.ceil(lz / len(fs)))[
            :lz
        ]  # This only works if len(fs) = len(zs)
        # TODO: Don't clip f. Practically fine for now.
    else:
        x, y, z, t, f = np.meshgrid(xs, ys, zs, thetas, fs)

    return np.vstack([x.ravel(), y.ravel(), z.ravel(), t.ravel(), f.ravel()]).T


def calc_num_tiles(dist, overlap, roi_length):
    """Calculate the number of tiles to divide a space dist along a single axis with an
    ROI of size roi_length and a fractional overlap between ROIs of overlap.

    Watch out! This has no indication of what the tiles should actually look like (no
    information about sign, etc.).

    Parameters
    ----------
    dist : float
        Total distance to tile with ROIs. A measure from the closed boundaries of the
        region to tile (e.g. left side of the first tile all the way to the right side
        of the last tile for x-dimension low -> high). Positive.
    overlap : float
        Fraction of roi_length that overlaps in each tile. Value between 0 and 1.
    roi_length : float
        The length of the ROI along this dimension. Positive.

    Returns
    -------
    num_tiles : int
        Number of tiles needed to cover this distance.

    """
    if dist == 0 or roi_length <= 0 or overlap >= 1:
        num_tiles = 1
    else:
        ov = overlap * roi_length  # True overlap in distance units
        num_tiles = ceil(
            (dist - ov) / (roi_length - ov)
        )  # ceil(abs(dist - ov) / abs(roi_length - ov))

    return int(num_tiles)


def update_table(table, pos, append=False):
    """Updates and redraws table based on given list.

    List is converted to a pandas dataframe before setting data in table.

    Parameters
    ----------
    table: Multi_Position_Table object
        Instance of multiposition table in GUI
    pos: list or np.array
        List or np.array of positions to be added to table. Each row contains an X, Y,
        Z, R, F position
    append: bool
        Append the new positions to the table

    Returns
    -------
    None :
        Table is updated
    """
    frame = pd.DataFrame(pos, columns=list("XYZRF"))
    if append:
        table.model.df = table.model.df.append(frame, ignore_index=True)
    else:
        table.model.df = frame
    table.currentrow = table.model.df.shape[0] - 1
    table.update_rowcolors()
    table.redraw()
    table.tableChanged()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import tkinter as tk
from tkinter import ttk

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox, ValidatedEntry

# Logging Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class TilingWizardPopup:
    """Popup for tiling parameters in View."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the Tile Wizard Popup

        Parameters
        ----------
        root : tk.Tk
            The root window.
        *args : list
            Arguments.
        **kwargs : dict
            Keyword arguments.
        """

        #: PopUp: The popup window.
        self.popup = PopUp(
            root,
            "Multi-Position Tiling Wizard",
            "630x420+330+330",
            top=False,
            transient=False,
        )

        # Storing the content frame of the popup, this will be the parent of
        # the widgets
        content_frame = self.popup.get_frame()
        content_frame.columnconfigure(0, pad=5)
        content_frame.columnconfigure(1, pad=5)
        content_frame.rowconfigure(0, pad=5)
        content_frame.rowconfigure(1, pad=5)
        content_frame.rowconfigure(2, pad=5)

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        # Sub Frames
        action_buttons = ttk.Frame(content_frame, padding=(0, 5, 0, 0))
        pos_grid = ttk.Frame(content_frame, padding=(0, 5, 0, 0))
        data = ttk.Frame(content_frame, padding=(0, 5, 0, 0))

        action_buttons.grid(row=0, sticky=tk.NSEW)
        pos_grid.grid(row=1, sticky=tk.NSEW)
        data.grid(row=2, sticky=tk.NSEW)

        #: dict: Dictionary for all the variables
        self.inputs = {}
        #: dict: Dictionary for all the buttons
        self.buttons = {}

        names = [
            "set_table",
            "x_start",
            "x_end",
            "y_start",
            "y_end",
            "z_start",
            "z_end",
        ]

        entry_names = ["x_dist", "x_tiles", "y_dist", "y_tiles", "z_dist", "z_tiles"]

        dist_labels = [
            "X Distance",
            "Num. Tiles",
            "Y Distance",
            "Num. Tiles",
            "Z Distance",
            "Num. Tiles",
        ]

        # Action buttons
        btn_labels = [
            "Populate Multi-Position Table",
            "Set X Start",
            "Set X End",
            "Set Y Start",
            "Set Y End",
            "Set Z Start",
            "Set Z End",
        ]

        # Position buttons
        for i in range(len(names)):
            if i == 0:
                self.buttons[names[i]] = ttk.Button(action_buttons, text=btn_labels[i])
                self.buttons[names[i]].grid(
                    row=0, column=i, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
                )
            if i > 0:
                self.buttons[names[i]] = ttk.Button(pos_grid, text=btn_labels[i])
                self.buttons[names[i]].grid(
                    row=i - 1, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
                )

                # Validated Spinbox
                self.inputs[names[i]] = LabelInput(
                    parent=pos_grid,
                    input_class=ValidatedSpinbox,
                    input_var=tk.StringVar(),
                )
                self.inputs[names[i]].grid(
                    row=i - 1, column=1, sticky=tk.NSEW, pady=(20, 0), padx=5
                )
                self.inputs[names[i]].widget.state(["disabled"])

        # Dist and Tile entries
        for i in range(len(entry_names)):
            self.inputs[entry_names[i]] = LabelInput(
                parent=pos_grid,
                label=dist_labels[i],
                input_class=ValidatedEntry,
                input_var=tk.StringVar(),
            )
            self.inputs[entry_names[i]].grid(
                row=i, column=2, sticky=tk.NSEW, padx=(5, 0), pady=(17, 0)
            )
            self.inputs[entry_names[i]].widget.state(["disabled"])

        self.inputs["percent_overlay"] = LabelInput(
            parent=data,
            label="Percent Overlay",
            input_class=ValidatedSpinbox,
            input_var=tk.StringVar(),
            input_args={"width": 5, "increment": 5, "from_": 0, "to": 100},
        )
        self.inputs["percent_overlay"].grid(
            row=0, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
        )

        self.inputs["total_tiles"] = LabelInput(
            parent=data,
            label="Total Tiles",
            input_class=ValidatedEntry,
            input_var=tk.StringVar(),
        )
        self.inputs["total_tiles"].widget.state(["disabled"])
        self.inputs["total_tiles"].grid(
            row=0, column=2, sticky=tk.NSEW, padx=5, pady=(5, 0)
        )

        # Formatting
        self.inputs["total_tiles"].grid(padx=(160, 0))

    # Getters
    def get_variables(self):
        """Get the variables tied to the widgets

        This function returns a dictionary of all the variables that are tied to each
        widget name. The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            A dictionary of all the variables that are tied to each widget name.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Get the widgets

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            A dictionary of all the widgets that are tied to each widget name.
        """
        return self.inputs

    def get_buttons(self):
        """Get the buttons

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        self.buttons : dict
            A dictionary of all the buttons that are tied to each button name.
        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class GUIController:
    """Base class for GUI controllers"""

    def __init__(self, view, parent_controller=None):
        """Initializes GUI controller

        Parameters
        ----------
        view : tkinter.Tk
            GUI view
        parent_controller : Controller
            parent controller
        """
        #: tkinter.Tk: GUI view
        self.view = view
        #: Controller: parent controller
        self.parent_controller = parent_controller

    def initialize(self):
        """This function is called when the controller is initialized

        This function initializes GUI based on configuration setting parameter:
        configuration_controller set range value for entry or spinbox widgets;
        add values to combobox get other necessary information for configuration.yml
        """
        pass

    def set_experiment_values(self):
        """Sets values of widgets based on experiment setting

        setting_dict is a dictionary
        """
        pass

    def update_experiment_values(self):
        """Collects all the values of widgets

        setting_dict is a reference of experiment dictionary
        update the dictionary directly
        """
        pass

    def execute(self, command, *args):
        """This function is called when a command is passed from a child process.

        Parameters
        ----------
        command : str
            command name
        args : tuple
            command arguments
        """

        self.show_verbose_info("command passed from child:", command)
        pass

    def show_verbose_info(self, *info):
        """Prints verbose information to the console

        Parameters
        ----------
        info : tuple
            information to be printed
        """
        logger.info(f"{self.__class__.__name__} : {info}")



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
from functools import reduce
import datetime

# Third Party Imports
import numpy as np
import tkinter as tk

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.controller.sub_controllers.channel_setting_controller import (
    ChannelSettingController,
)
from navigate.controller.sub_controllers.tiling_wizard_controller2 import (
    TilingWizardController,
)

# View Imports that are not called on startup
from navigate.view.popups.tiling_wizard_popup2 import TilingWizardPopup

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ChannelsTabController(GUIController):
    """Controller for the channels tab in the main window."""

    def __init__(self, view, parent_controller=None):
        """Initialize the ChannelsTabController.

        Parameters
        ----------
        view : tkinter.Frame
            The view that this controller will control.
        parent_controller : navigate.controller.main_controller.MainController
            The parent controller of this controller.
        """
        super().__init__(view, parent_controller)

        #: bool: Whether or not the user has selected to save the data.
        self.is_save = False

        #: str: The current acquisition mode.
        self.mode = "stop"

        #: bool: Whether the controller is in the initialization phase.
        self.in_initialization = True

        # sub-controllers
        self.channel_setting_controller = ChannelSettingController(
            self.view.channel_widgets_frame,
            self,
            self.parent_controller.configuration_controller,
        )

        self.view.stack_timepoint_frame.exp_time_spinbox.set_precision(0)
        # Get Widgets and Buttons from stack_acquisition_settings in view
        #: dict: The widgets in the stack acquisition settings frame.
        self.stack_acq_widgets = self.view.stack_acq_frame.get_widgets()
        #: dict: The values in the stack acquisition settings frame.
        self.stack_acq_vals = self.view.stack_acq_frame.get_variables()
        #: dict: The buttons in the stack acquisition settings frame.
        self.stack_acq_buttons = self.view.stack_acq_frame.get_buttons()

        # stack acquisition event binds
        self.stack_acq_vals["step_size"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["start_position"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["end_position"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["start_focus"].trace_add(
            "write", self.update_z_steps
        )  # TODO: could be remove later
        self.stack_acq_buttons["set_start"].configure(
            command=self.update_start_position
        )
        self.stack_acq_buttons["set_end"].configure(command=self.update_end_position)

        # stack acquisition_variables
        #: float: The z origin of the stack.
        self.z_origin = 0
        #: float: The focus origin of the stack.
        self.focus_origin = 0
        #: float: The stage velocity.
        self.stage_velocity = None
        #: float: The filter wheel delay.
        self.filter_wheel_delay = None
        #: dict: The microscope state dictionary.
        self.microscope_state_dict = None

        # laser/stack cycling event binds
        self.stack_acq_vals["cycling"].trace_add("write", self.update_cycling_setting)

        # time point setting variables
        #: dict: Dictionary of time point settings.
        self.timepoint_vals = {
            "is_save": self.view.stack_timepoint_frame.save_data,
            "timepoints": self.view.stack_timepoint_frame.exp_time_spinval,
            "stack_acq_time": self.view.stack_timepoint_frame.stack_acq_spinval,
            "stack_pause": self.view.stack_timepoint_frame.stack_pause_spinval,
            "experiment_duration": self.view.stack_timepoint_frame.total_time_spinval,
        }
        self.timepoint_vals[
            "timepoint_interval"
        ] = self.view.stack_timepoint_frame.timepoint_interval_spinval

        # timepoint event binds
        self.timepoint_vals["is_save"].trace_add("write", self.update_save_setting)
        self.timepoint_vals["timepoints"].trace_add(
            "write", lambda *args: self.update_timepoint_setting(True)
        )
        self.timepoint_vals["stack_pause"].trace_add(
            "write", lambda *args: self.update_timepoint_setting(True)
        )

        # Multi Position Acquisition
        #: bool: Whether or not the user has selected to use multiposition.
        self.is_multiposition = False
        #: bool: cache multiposition flag
        self.is_multiposition_cache = False
        #: bool: Whether or not the user has selected to use multiposition.
        self.is_multiposition_val = self.view.multipoint_frame.on_off
        self.view.multipoint_frame.save_check.configure(
            command=self.toggle_multiposition
        )
        self.view.multipoint_frame.buttons["tiling"].configure(
            command=self.launch_tiling_wizard
        )

        # Waveform Parameters
        self.view.quick_launch.buttons["waveform_parameters"].configure(
            command=self.launch_waveform_parameters
        )

        # Autofocus Settings
        self.view.quick_launch.buttons["autofocus_button"].configure(
            command=self.launch_autofocus_settings
        )

        self.initialize()

    def initialize(self):
        """Initializes widgets and gets other necessary configuration

        Examples
        --------
        >>> self.initialize()
        """
        config = self.parent_controller.configuration_controller

        self.stack_acq_widgets["cycling"].widget["values"] = ["Per Z", "Per Stack"]
        self.stage_velocity = config.stage_setting_dict["velocity"]
        self.filter_wheel_delay = config.filter_wheel_setting_dict["filter_wheel_delay"]
        self.channel_setting_controller.initialize()
        self.set_spinbox_range_limits(config.configuration["configuration"]["gui"])
        self.show_verbose_info("channels tab has been initialized")

    def populate_experiment_values(self):
        """Distribute initial MicroscopeState values to this and sub-controllers and
        associated views.

        Examples
        --------
        >>> self.populate_experiment_values()
        """
        self.in_initialization = True
        self.microscope_state_dict = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]
        if self.microscope_state_dict["step_size"] < 0:
            self.microscope_state_dict["step_size"] = -self.microscope_state_dict[
                "step_size"
            ]
        self.set_info(self.stack_acq_vals, self.microscope_state_dict)
        # self.set_info(self.conpro_acq_vals, self.microscope_state_dict)
        self.set_info(self.timepoint_vals, self.microscope_state_dict)

        # check configuration for multiposition settings
        self.is_multiposition_val.set(self.microscope_state_dict["is_multiposition"])
        self.is_multiposition_cache = self.is_multiposition
        self.toggle_multiposition()

        # validate
        self.view.stack_timepoint_frame.stack_pause_spinbox.trigger_focusout_validation()
        self.view.stack_timepoint_frame.exp_time_spinbox.trigger_focusout_validation()

        if self.microscope_state_dict["stack_cycling_mode"] not in [
            "per_z",
            "per_stack",
        ]:
            self.microscope_state_dict["stack_cycling_mode"] = "per_stack"
        self.stack_acq_vals["cycling"].set(
            "Per Z"
            if self.microscope_state_dict["stack_cycling_mode"] == "per_z"
            else "Per Stack"
        )
        if self.microscope_state_dict["conpro_cycling_mode"] not in [
            "per_plane",
            "per_stack",
        ]:
            self.microscope_state_dict["conpro_cycling_mode"] = "per_plane"
        self.channel_setting_controller.populate_experiment_values(
            self.microscope_state_dict["channels"]
        )

        # after initialization
        self.in_initialization = False
        self.channel_setting_controller.in_initialization = False
        self.update_z_steps()

        self.show_verbose_info("Channels tab has been set new values")

    def set_spinbox_range_limits(self, settings):
        """This function will set the spinbox widget's values of from_, to, step

        Parameters
        ----------
        settings : dict
            dictionary of settings from configuration file

        Examples
        --------
        >>> self.set_spinbox_range_limits(settings)
        """
        temp_dict = {
            self.stack_acq_widgets["step_size"]: settings["stack_acquisition"][
                "step_size"
            ],
            self.stack_acq_widgets["start_position"]: settings["stack_acquisition"][
                "start_pos"
            ],
            self.stack_acq_widgets["end_position"]: settings["stack_acquisition"][
                "end_pos"
            ],
            self.view.stack_timepoint_frame.stack_pause_spinbox: settings["timepoint"][
                "stack_pause"
            ],
            self.view.stack_timepoint_frame.exp_time_spinbox: settings["timepoint"][
                "timepoints"
            ],
        }
        for idx, widget in enumerate(temp_dict):
            # Hacky Solution until stack time points are converted to LabelInput
            if idx < 3:
                widget.widget.configure(from_=temp_dict[widget]["min"])
                widget.widget.configure(to=temp_dict[widget]["max"])
                widget.widget.configure(increment=temp_dict[widget]["step"])
            else:
                widget.configure(from_=temp_dict[widget]["min"])
                widget.configure(to=temp_dict[widget]["max"])
                widget.configure(increment=temp_dict[widget]["step"])

        # channels setting
        self.channel_setting_controller.set_spinbox_range_limits(settings["channels"])

    def set_mode(self, mode):
        """Change acquisition mode.

        Parameters
        ----------
        mode : str
            acquisition mode

        Examples
        --------
        >>> self.set_mode(mode)
        """
        image_mode = self.microscope_state_dict["image_mode"]
        self.mode = mode
        self.channel_setting_controller.set_mode(mode)

        state_readonly = "readonly" if mode == "stop" else "disabled"
        if mode != "stop":
            state = "disabled"
        elif image_mode == "live" or image_mode == "single":
            state = "disabled"
        else:
            state = "normal"
        for widget_name in [
            "start_position",
            "start_focus",
            "end_position",
            "end_focus",
            "step_size",
        ]:
            self.stack_acq_widgets[widget_name].widget["state"] = state
        self.stack_acq_widgets["cycling"].widget["state"] = (
            "readonly" if state == "normal" else "disabled"
        )
        self.view.stack_timepoint_frame.save_check["state"] = (
            "normal" if image_mode == "single" and mode == "stop" else state
        )
        self.view.stack_timepoint_frame.stack_pause_spinbox["state"] = state
        self.view.stack_timepoint_frame.exp_time_spinbox["state"] = state

        # multi-position flag
        if mode == "stop":
            self.is_multiposition_val.set(self.is_multiposition_cache)
            self.toggle_multiposition()
        else:
            self.is_multiposition_cache = self.is_multiposition
            if mode == "customized":
                self.is_multiposition_val.set(False)
                self.toggle_multiposition()

        if image_mode == "customized" or mode != "stop":
            self.disable_multiposition_btn()
        else:
            self.enable_multiposition_btn()

        self.show_verbose_info("acquisition mode has been changed to", mode)

    def update_z_steps(self, *args):
        """Recalculates the number of slices that will be acquired in a z-stack.

        Requires GUI to have start position, end position, or step size changed.
        Sets the number of slices in the model and the GUI.
        Sends the current values to central/parent controller

        Parameters
        ----------
        args : dict
            Values is a dict as follows {'step_size':  'start_position': ,
                                         'end_position': ,'number_z_steps'}

        Examples
        --------
        >>> self.update_z_steps()
        """

        # won't do any calculation when initialization
        if self.in_initialization:
            return

        # Calculate the number of slices and set GUI
        try:
            # validate the spinbox's value
            start_position = float(self.stack_acq_vals["start_position"].get())
            end_position = float(self.stack_acq_vals["end_position"].get())
            step_size = float(self.stack_acq_vals["step_size"].get())
            if step_size < 0.001:
                self.stack_acq_vals["number_z_steps"].set(0)
                self.stack_acq_vals["abs_z_start"].set(0)
                self.stack_acq_vals["abs_z_end"].set(0)
                return
        except tk._tkinter.TclError:
            self.stack_acq_vals["number_z_steps"].set(0)
            self.stack_acq_vals["abs_z_start"].set(0)
            self.stack_acq_vals["abs_z_end"].set(0)
            return
        except (KeyError, AttributeError):
            logger.error("Error caught: updating z_steps")
            return

        # if step_size < 0.001:
        #     step_size = 0.001
        #     self.stack_acq_vals['step_size'].set(step_size)

        number_z_steps = int(
            np.ceil(np.abs((end_position - start_position) / step_size))
        )
        self.stack_acq_vals["number_z_steps"].set(number_z_steps)

        # Shift the start/stop positions by the relative position
        flip_flags = self.parent_controller.configuration_controller.stage_flip_flags
        if flip_flags["z"]:
            self.stack_acq_vals["abs_z_start"].set(self.z_origin + end_position)
            self.stack_acq_vals["abs_z_end"].set(self.z_origin + start_position)
        else:
            self.stack_acq_vals["abs_z_start"].set(self.z_origin + start_position)
            self.stack_acq_vals["abs_z_end"].set(self.z_origin + end_position)

        # update experiment MicroscopeState dict
        self.microscope_state_dict["start_position"] = start_position
        self.microscope_state_dict["end_position"] = end_position
        self.microscope_state_dict["step_size"] = step_size * (
            -1 if flip_flags["z"] else 1
        )
        self.microscope_state_dict["number_z_steps"] = number_z_steps
        self.microscope_state_dict["abs_z_start"] = self.stack_acq_vals[
            "abs_z_start"
        ].get()
        self.microscope_state_dict["abs_z_end"] = self.stack_acq_vals["abs_z_end"].get()
        try:
            self.microscope_state_dict["start_focus"] = self.stack_acq_vals[
                "start_focus"
            ].get()
        except tk._tkinter.TclError:
            self.microscope_state_dict["start_focus"] = 0
        try:
            self.microscope_state_dict["end_focus"] = self.stack_acq_vals[
                "end_focus"
            ].get()
        except tk._tkinter.TclError:
            self.microscope_state_dict["end_focus"] = 0
        self.microscope_state_dict["stack_z_origin"] = self.z_origin
        self.microscope_state_dict["stack_focus_origin"] = self.focus_origin

        self.update_timepoint_setting()
        self.show_verbose_info(
            "stack acquisition settings on channels tab have been changed and "
            "recalculated"
        )

    def update_start_position(self, *args):
        """Get new z starting position from current stage parameters.

        Parameters
        ----------
        args : dict
            Values is a dict as follows {'start_position': , 'abs_z_start': ,
            'stack_z_origin': }

        Examples
        --------
        >>> self.update_start_position()
        """

        # We have a new origin
        self.z_origin = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["z"]
        self.focus_origin = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["f"]

        flip_flags = self.parent_controller.configuration_controller.stage_flip_flags
        if flip_flags["z"]:
            self.stack_acq_vals["end_position"].set(0)
            self.stack_acq_vals["end_focus"].set(0)
        else:
            self.stack_acq_vals["start_position"].set(0)
            self.stack_acq_vals["start_focus"].set(0)

        # Propagate parameter changes to the GUI
        self.update_z_steps()

    def update_end_position(self, *args):
        """Get new z ending position from current stage parameters

        Parameters
        ----------
        args : dict
            Values is a dict as follows {'end_position': , 'abs_z_end': ,
            'stack_z_origin': }

        Examples
        --------
        >>> self.update_end_position()
        """
        # Grab current values
        z_end = self.parent_controller.configuration["experiment"]["StageParameters"][
            "z"
        ]
        focus_end = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["f"]

        z_start = self.z_origin
        focus_start = self.focus_origin

        if z_end < z_start:
            # Sort so we are always going low to high
            z_start, z_end = z_end, z_start
            focus_start, focus_end = focus_end, focus_start

        # set origin to be in the middle of start and end
        self.z_origin = (z_start + z_end) / 2
        self.focus_origin = (focus_start + focus_end) / 2

        # Propagate parameter changes to the GUI
        flip_flags = self.parent_controller.configuration_controller.stage_flip_flags
        start_pos = z_start - self.z_origin
        end_pos = z_end - self.z_origin
        start_focus = focus_start - self.focus_origin
        end_focus = focus_end - self.focus_origin
        if flip_flags["z"]:
            start_pos, end_pos = end_pos, start_pos
            start_focus, end_focus = end_focus, start_focus
        self.stack_acq_vals["start_position"].set(start_pos)
        self.stack_acq_vals["start_focus"].set(start_focus)
        self.stack_acq_vals["end_position"].set(end_pos)
        self.stack_acq_vals["end_focus"].set(end_focus)
        self.update_z_steps()

    def update_cycling_setting(self, *args):
        """Update the cycling settings in the model and the GUI.

        You can collect different channels in different formats.
        In the perZ format: Slice 0/Ch0, Slice0/Ch1, Slice1/Ch0, Slice1/Ch1, etc
        in the perStack format: Slice 0/Ch0, Slice1/Ch0... SliceN/Ch0.  Then it repeats
        with Ch1

        Parameters
        ----------
        args : dict
            Values is a dict as follows {'cycling_setting': , 'cycling_setting': ,
                                         'stack_z_origin': }

        Examples
        --------
        >>> self.update_cycling_setting()
        """
        # won't do any calculation when initializing
        if self.in_initialization:
            return
        # update experiment MicroscopeState dict
        self.microscope_state_dict["stack_cycling_mode"] = (
            "per_stack"
            if self.stack_acq_vals["cycling"].get() == "Per Stack"
            else "per_z"
        )

        # recalculate time point settings
        self.update_timepoint_setting()

        self.show_verbose_info("Cycling setting on channels tab has been changed")

    def update_save_setting(self, *args):
        """Tell the centrol/parent controller 'save_data' is selected.

        Does not do any calculation when initializing the software.

        Parameters
        ----------
        args : dict
            Values is a dict as follows {'save_data': , 'save_data': ,
                                         'stack_z_origin': }

        Examples
        --------
        >>> self.update_save_setting()
        """

        if self.in_initialization:
            return
        self.is_save = self.timepoint_vals["is_save"].get()
        # update experiment MicroscopeState dict
        self.microscope_state_dict["is_save"] = self.is_save
        self.parent_controller.execute("set_save", self.is_save)
        self.show_verbose_info("Save data option has been changed to", self.is_save)

    def update_timepoint_setting(self, call_parent=False):
        """Automatically calculates the stack acquisition time based on the number of
        time points, channels, and exposure time.

        TODO: Add necessary computation for 'Stack Acq.Time', 'Timepoint Interval',
        'Experiment Duration'?

        Does not do any calculation when initializing the software.
        Order of priority for perStack: timepoints > positions > channels > z-steps
                                        > delay
        ORder of priority for perZ: timepoints > positions > z-steps > delays > channels

        Parameters
        ----------
        call_parent : bool
            Tell parent controller that time point setting has changed.

        Examples
        --------
        >>> self.update_timepoint_setting()
        """

        if self.in_initialization:
            return
        channel_settings = self.microscope_state_dict["channels"]
        number_of_positions = (
            self.parent_controller.multiposition_tab_controller.get_position_num()
            if self.is_multiposition
            else 1
        )
        channel_exposure_time = []
        # validate the spinbox's value
        try:
            number_of_timepoints = int(float(self.timepoint_vals["timepoints"].get()))
            number_of_slices = int(self.stack_acq_vals["number_z_steps"].get())
            for channel_id in channel_settings.keys():
                channel = channel_settings[channel_id]
                if channel["is_selected"]:
                    channel_exposure_time.append(float(channel["camera_exposure_time"]))
            if len(channel_exposure_time) == 0:
                return
        except (tk._tkinter.TclError, ValueError):
            self.timepoint_vals["experiment_duration"].set("0")
            self.timepoint_vals["stack_acq_time"].set("0")
            return
        except (KeyError, AttributeError):
            logger.error("Error caught: updating timepoint setting")
            return

        perStack = self.stack_acq_vals["cycling"].get() == "Per Stack"

        # Initialize variable to keep track of how long the entire experiment will take.
        # Includes time, positions, channels...
        experiment_duration = 0

        # Initialize variable to calculate how long it takes to acquire a single volume
        # for all of the channels. Only calculate once at the beginning.
        stack_acquisition_duration = 0

        for position_idx in range(number_of_positions):
            # For multiple positions, need to account for the time necessary to move
            # the stages that distance. In theory, these positions would be
            # populated in that 'pandastable' or some other data structure.

            # Determine the largest distance to travel between positions.  Assume
            # all axes move the same velocity This assumes that we are in a
            # multi-position mode. Not yet implemented.
            # x1, y1, z1, theta1, f1, = position_start.values()
            # x2, y2, z1, theta2, f1 = position_end.values()
            # distance = [x2-x1, y2-y1, z2-z1, theta2-theta1, f2-f1]
            # max_distance_idx = np.argmax(distance)
            # Now if we are going to do this properly, we would need to do this for
            # all of the positions so that we can calculate the total experiment
            # time. Probably assemble a matrix of all the positions and then do
            # the calculations.

            stage_delay = 0  # distance[max_distance_idx]/self.stage_velocity
            # TODO False value.

            # If we were actually acquiring the data, we would call the function to
            # move the stage here.
            experiment_duration = experiment_duration + stage_delay

            for channel_idx in range(len(channel_exposure_time)):
                if perStack:
                    # In the perStack mode, we only need to account for the time
                    # necessary for the filter wheel to change between each
                    # image stack.
                    if channel_idx == 0 and position_idx == 0:
                        stack_acquisition_duration += (
                            channel_exposure_time[channel_idx] / 1000 * number_of_slices
                        )
                else:
                    if position_idx == 0:
                        stack_acquisition_duration += (
                            channel_exposure_time[channel_idx] / 1000 * number_of_slices
                        )

                experiment_duration += (
                    channel_exposure_time[channel_idx] / 1000 * number_of_slices
                )

            try:
                stack_pause = float(self.timepoint_vals["stack_pause"].get())
            except ValueError:
                stack_pause = 0
            experiment_duration = experiment_duration + stack_pause
        experiment_duration *= number_of_timepoints

        # Change the filter wheel here before the start of the acquisition.
        if len(channel_exposure_time) > 1:
            filter_wheel_change_times = number_of_timepoints * (
                1 if perStack else number_of_slices
            )
            experiment_duration += self.filter_wheel_delay * filter_wheel_change_times
        else:
            experiment_duration += self.filter_wheel_delay
        self.timepoint_vals["experiment_duration"].set(
            str(datetime.timedelta(seconds=experiment_duration))
        )
        self.timepoint_vals["stack_acq_time"].set(
            str(datetime.timedelta(seconds=stack_acquisition_duration))
        )

        # update experiment MicroscopeState dict
        self.microscope_state_dict["timepoints"] = number_of_timepoints
        self.microscope_state_dict["stack_pause"] = self.timepoint_vals[
            "stack_pause"
        ].get()
        # self.microscope_state_dict['timepoint_interval']
        self.microscope_state_dict["stack_acq_time"] = stack_acquisition_duration
        self.microscope_state_dict["experiment_duration"] = experiment_duration

        self.show_verbose_info(
            "timepoint settings on channels tab have been changed and recalculated"
        )

    def toggle_multiposition(self):
        """Toggle Multi-position Acquisition.

        Recalculates the experiment duration.

        Examples
        --------
        >>> self.toggle_multiposition()
        """
        self.is_multiposition = self.is_multiposition_val.get()
        self.microscope_state_dict["is_multiposition"] = self.is_multiposition
        self.update_timepoint_setting()
        self.show_verbose_info("Multi-position:", self.is_multiposition)

    def disable_multiposition_btn(self):
        """Disable multiposition button"""
        self.view.multipoint_frame.save_check.config(state="disabled")

    def enable_multiposition_btn(self):
        """Enable multiposition button"""
        self.view.multipoint_frame.save_check.config(state="normal")

    def launch_waveform_parameters(self):
        """Launches waveform parameters popup."""
        self.parent_controller.menu_controller.popup_waveform_setting()

    def launch_autofocus_settings(self):
        """Launches autofocus settings popup."""
        self.parent_controller.menu_controller.popup_autofocus_setting()

    def launch_tiling_wizard(self):
        """Launches tiling wizard popup.

        Will only launch when button in GUI is pressed, and will not duplicate.
        Pressing button again brings popup to top

        Examples
        --------
        >>> self.launch_tiling_wizard()
        """

        if hasattr(self, "tiling_wizard_controller"):
            self.tiling_wizard_controller.showup()
            return
        tiling_wizard = TilingWizardPopup(self.view)
        self.tiling_wizard_controller = TilingWizardController(tiling_wizard, self)

    def set_info(self, vals, values):
        """Set values to a list of variables.

        Parameters
        ----------
        vals : list
            List of variables to set.
        values : list
            List of values to set to variables.

        Examples
        --------
        >>> self.set_info([self.timepoint_vals['timepoint_interval'],
                           self.timepoint_vals['stack_pause']], [1, 1])
        """
        for name in values.keys():
            if name in vals:
                vals[name].set(values[name])

    def execute(self, command, *args):
        """Execute Command in the parent controller.

        Parameters
        ----------
        command : str
            recalculate_timepoint, channel, move_stage_and_update_info,
            get_stage_position
        args : list
            List of arguments to pass to the command.

        Returns
        -------
        command : object
            Returns parent_controller.execute(command) if command = 'get_stage_position'

        Examples
        --------
        >>> self.execute('recalculate_timepoint')
        """
        if command == "recalculate_timepoint":
            # update selected channels num
            self.microscope_state_dict["selected_channels"] = reduce(
                lambda count, channel: count + (channel["is_selected"] is True),
                self.microscope_state_dict["channels"].values(),
                0,
            )
            self.update_timepoint_setting()
            # update framerate info in camera setting tab
            exposure_time = max(
                map(
                    lambda channel: float(channel["camera_exposure_time"])
                    if channel["is_selected"]
                    else 0,
                    self.microscope_state_dict["channels"].values(),
                )
            )
            self.parent_controller.camera_setting_controller.update_exposure_time(
                exposure_time
            )
        elif (command == "channel") or (command == "update_setting"):
            self.view.after(
                1000, lambda: self.parent_controller.execute(command, *args)
            )

        self.show_verbose_info("Received command from child", command, args)

    def verify_experiment_values(self):
        """Verify channel tab settings and return warning info

        Returns
        -------
        string
            Warning info
        """
        warning = self.channel_setting_controller.verify_experiment_values()
        if warning:
            return warning
        if self.microscope_state_dict["image_mode"] not in ["live", "single"]:
            if (
                self.microscope_state_dict["number_z_steps"]
                != self.stack_acq_vals["number_z_steps"].get()
            ):
                return "There is something wrong with the stack settings!"
            if self.microscope_state_dict["number_z_steps"] < 1:
                return "The number of Z steps should be at least 1!"
            try:
                float(self.microscope_state_dict["stack_pause"])
            except:
                return "Stack pause should be a valid number!"
            if self.microscope_state_dict["timepoints"] < 1:
                return "Timepoints should be at least 1!"
        return None



"""change_settings_test.py

Note: Written for Prime 2020; some settings may not be valid for other cameras!
Change settings in the `camera_settings.py` file.
"""
try:
    from pyvcam import pvc
    from pyvcam.camera import Camera
    from camera_settings import apply_settings

    def print_settings(camera):
        print("clear mode: {}".format(camera.clear_modes[camera.clear_mode]))
        print("exposure mode: {}".format(camera.exp_modes[camera.exp_mode]))
        print("readout port: {}".format(camera.readout_port))
        print("speed table index: {}".format(camera.speed_table_index))
        print("gain: {}".format(camera.gain))

    # Initialize PVCAM
    pvc.init_pvcam()

    # Find the first available camera.
    camera = next(Camera.detect_camera())
    camera.open()

    # Show camera name and speed table.
    print(camera)

    print("\nBefore changing settings:")
    print_settings(camera)

    # Change the camera settings from the separate file.
    apply_settings(camera)

    print("\nAfter changing settings:")
    print_settings(camera)

    # Cleanup before exit
    camera.close()
    pvc.uninit_pvcam()
except ModuleNotFoundError:
    print("Install pyvcam.")



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import logging
import time

# Third Party Imports

# Local Imports
from navigate.model.devices.stages.stage_base import StageBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class SyntheticStage(StageBase):
    """Synthetically generated stage for testing purposes."""

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the stage.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : str
            Connection string for the device.
        configuration : dict
            Configuration dictionary for the device.
        device_id : int
            Device ID for the device.
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        #: float: The default speed of the stage in mm/s.
        self.default_speed = 7.68 * 0.67

        # Default axes mapping
        axes_mapping = {"x": "X", "y": "Y", "z": "Z", "theta": "Theta", "f": "F"}
        if not self.axes_mapping:
            #: dict: Dictionary mapping software axes to hardware axes.
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }

        self.sample_rate = 10000
        self.volts_per_micron = "0.1 * x"
        self.remote_focus_delay = 0.05
        self.camera_delay_percent = 0.01
        self.remote_focus_ramp_falling = 0.1

    def report_position(self):
        """Report the current position of the stage.

        Returns
        -------
        position_dict : dict
            Dictionary containing the current position of the stage.
        """
        return self.get_position_dict()

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Implement movement logic along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        axis_abs = self.get_abs_position(axis, abs_pos)
        if axis_abs == -1e50:
            return False

        if wait_until_done:
            time.sleep(0.025)

        # Move the stage
        setattr(self, f"{axis}_pos", axis_abs)
        return True

    def move_absolute(self, move_dictionary, wait_until_done=False):
        """Move stage along a single axis.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for one
            or more axes. Expects values in micrometers, except for theta, which is
            in degrees.
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        success : bool
            Was the move successful?
        """
        abs_pos_dict = self.verify_abs_position(move_dictionary)
        if not abs_pos_dict:
            return False

        for axis in abs_pos_dict:
            setattr(self, f"{axis}_pos", abs_pos_dict[axis])
        if wait_until_done is True:
            time.sleep(0.025)

        return True

    def load_sample(self):
        """Load a sample.

        Warning
        -------
            Not implemented.
        """
        self.y_pos = self.y_load_position

    def unload_sample(self):
        """Unload a sample.

        Warning
        -------
            Not implemented.
        """
        self.y_pos = self.y_unload_position

    def get_axis_position(self, axis):
        """Get the current position of the stage along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.

        Returns
        -------
        float
            The current position of the stage along the specified axis.
        """
        return getattr(self, f"{axis}_pos")

    def set_speed(self, velocity_dict):
        """Set the speed of the stage.

        Parameters
        ----------
        velocity_dict : dict
            Dictionary containing the speed of the stage along each axis.
        """
        pass

    def get_speed(self, axis):
        """Get the speed of the stage.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.

        Returns
        -------
        float
            The speed of the stage along the specified axis.
        """

        return 1

    def scanr(self, start_position_mm, end_position_mm, enc_divide, axis="z"):
        """Scan the stage using the constant velocity mode along a single axis.

        Parameters
        ----------
        start_position_mm : float
            Starting position of the scan in mm.
        end_position_mm : float
            Ending position of the scan in mm.
        enc_divide : int
            Encoder divide value.
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.

        """

        pass

    def start_scan(self, axis):
        """Start a scan along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.

        """
        pass

    def stop_scan(self):
        """Stop a scan."""
        pass

    def update_waveform(self, waveform_dict):
        print("*** update waveform:", waveform_dict.keys())
        pass



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging

# Third Party Imports

# Local Imports
from navigate.model.waveforms import sawtooth, sine_wave

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class GalvoBase:
    """GalvoBase Class - Parent class for galvanometers."""

    def __init__(self, microscope_name, device_connection, configuration, galvo_id=0):
        """Initialize the GalvoBase class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : dict
            Dictionary of device connections.
        configuration : dict
            Dictionary of configuration parameters.
        galvo_id : int
            Galvo ID.
        """

        #: dict: Dictionary of microscope configuration parameters.
        self.configuration = configuration

        #: str: Name of the microscope.
        self.microscope_name = microscope_name

        #: str: Name of the galvo.
        self.galvo_name = "Galvo " + str(galvo_id)

        #: dict: Dictionary of device connections.
        self.device_config = configuration["configuration"]["microscopes"][
            microscope_name
        ]["galvo"][galvo_id]

        #: int: Sample rate.
        self.sample_rate = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sample_rate"]

        #: float: Sweep time.
        self.sweep_time = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sweep_time"]

        #: float: Camera delay percent.
        self.camera_delay_percent = configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]["delay_percent"]

        #: float: Galvo max voltage.
        self.galvo_max_voltage = self.device_config["hardware"]["max"]

        #: float: Galvo min voltage.
        self.galvo_min_voltage = self.device_config["hardware"]["min"]

        #: float: Percent galvo ramp rising percent.
        self.remote_focus_ramp_falling = configuration["configuration"]["microscopes"][
            microscope_name
        ]["remote_focus_device"]["ramp_falling_percent"]

        # Galvo Waveform Information
        #: str: Galvo waveform. Waveform or Sawtooth.
        self.galvo_waveform = self.device_config.get("waveform", "sawtooth")

        #: dict: Dictionary of galvo waveforms.
        self.waveform_dict = {}

    def __del__(self):
        """Destructor"""
        pass

    def adjust(self, exposure_times, sweep_times):
        """Adjust the galvo waveform to account for the camera readout time.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of camera exposure time in seconds on a per-channel basis.
            e.g., exposure_times = {"channel_1": 0.1, "channel_2": 0.2}
        sweep_times : dict
            Dictionary of acquisition sweep time in seconds on a per-channel basis.
            e.g., sweep_times = {"channel_1": 0.1, "channel_2": 0.2}

        Returns
        -------
        waveform_dict : dict
            Dictionary that includes the galvo waveforms on a per-channel basis.
        """
        self.waveform_dict = dict.fromkeys(self.waveform_dict, None)
        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        microscope_name = microscope_state["microscope_name"]
        zoom_value = microscope_state["zoom"]
        galvo_parameters = self.configuration["waveform_constants"]["galvo_constants"][
            self.galvo_name
        ][microscope_name][zoom_value]
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["sample_rate"]

        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:

                # Get the Waveform Parameters - Assumes ETL Delay < Camera Delay.
                # Should Assert.
                exposure_time = exposure_times[channel_key]
                self.sweep_time = sweep_times[channel_key]

                # galvo Parameters
                try:
                    galvo_amplitude = float(galvo_parameters.get("amplitude", 0))
                    galvo_offset = float(galvo_parameters.get("offset", 0))
                    galvo_frequency = (
                        float(galvo_parameters.get("frequency", 0)) / exposure_time
                    )
                except ValueError as e:
                    logger.error(
                        f"{e} waveform constants.yml doesn't have parameter "
                        f"amplitude/offset/frequency for {self.galvo_name}"
                    )
                    return

                # Calculate the Waveforms
                if self.galvo_waveform == "sawtooth":
                    self.waveform_dict[channel_key] = sawtooth(
                        sample_rate=self.sample_rate,
                        sweep_time=self.sweep_time,
                        frequency=galvo_frequency,
                        amplitude=galvo_amplitude,
                        offset=galvo_offset,
                        phase=(self.camera_delay_percent / 100) * exposure_time,
                    )
                elif self.galvo_waveform == "sine":
                    self.waveform_dict[channel_key] = sine_wave(
                        sample_rate=self.sample_rate,
                        sweep_time=self.sweep_time,
                        frequency=galvo_frequency,
                        amplitude=galvo_amplitude,
                        offset=galvo_offset,
                        phase=self.device_config["phase"],
                    )
                elif self.galvo_waveform == "halfsaw":
                    new_wave = sawtooth(
                        sample_rate=self.sample_rate,
                        sweep_time=self.sweep_time,
                        frequency=galvo_frequency,
                        amplitude=galvo_amplitude,
                        offset=galvo_offset,
                        phase=(self.camera_delay_percent / 100) * exposure_time,
                    )
                    half_samples = (
                        new_wave.argmax() if galvo_amplitude > 0 else new_wave.argmin()
                    )
                    new_wave[:half_samples] = -galvo_offset
                    self.waveform_dict[channel_key] = new_wave
                else:
                    print("Unknown Galvo waveform specified in configuration file.")
                    self.waveform_dict[channel_key] = None
                    continue
                self.waveform_dict[channel_key][
                    self.waveform_dict[channel_key] > self.galvo_max_voltage
                ] = self.galvo_max_voltage
                self.waveform_dict[channel_key][
                    self.waveform_dict[channel_key] < self.galvo_min_voltage
                ] = self.galvo_min_voltage

        return self.waveform_dict

    def turn_off(self):
        """Turn off the galvo."""
        pass


# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import tkinter as tk
from tkinter import ttk
import logging

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.hover import Hover
from navigate.view.custom_widgets.hovermixin import HoverButton
from navigate.view.custom_widgets.validation import ValidatedSpinbox, ValidatedCombobox
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ChannelsTab(tk.Frame):
    """Channels Tab for the Main Window

    This tab is used to set the channels for the stack acquisition.
    """

    def __init__(self, setntbk, *args, **kwargs):
        """Initialization of the Channels Tab

        Parameters
        ----------
        setntbk : tk.Notebook
            The notebook that this tab is added to
        *args : tuple
            Positional arguments for tk.Frame
        **kwargs : dict
            Keyword arguments for tk.Frame

        """
        # Init Frame
        tk.Frame.__init__(self, setntbk, *args, **kwargs)

        #: int: The index of the tab
        self.index = 0

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: ChannelCreator: The frame that holds the channel settings
        self.channel_widgets_frame = ChannelCreator(self)
        self.channel_widgets_frame.grid(
            row=0, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: StackAcquisitionFrame: The frame that holds the stack acquisition settings
        self.stack_acq_frame = StackAcquisitionFrame(self)
        self.stack_acq_frame.grid(
            row=1, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: StackTimePointFrame: The frame that holds the time settings
        self.stack_timepoint_frame = StackTimePointFrame(self)
        self.stack_timepoint_frame.grid(
            row=3, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: MultiPointFrame: The frame that holds the multipoint settings
        self.multipoint_frame = MultiPointFrame(self)
        self.multipoint_frame.grid(
            row=4, column=0, columnspan=1, sticky=tk.NSEW, padx=10, pady=10
        )

        #: QuickLaunchFrame: The frame that holds the quick launch buttons
        self.quick_launch = QuickLaunchFrame(self)
        self.quick_launch.grid(
            row=4, column=1, columnspan=2, sticky=tk.NSEW, padx=10, pady=10
        )



class ChannelCreator(ttk.Labelframe):
    """Channel Creator

    This frame is used to create the channels for the stack acquisition.
    """

    def __init__(self, channels_tab, *args, **kwargs):
        """Initialization of the Channel Creator

        Parameters
        ----------
        channels_tab : tk.Frame
            The frame that this frame is added to
        *args : tuple
            Positional arguments for ttk.Labelframe
        **kwargs : dict
            Keyword arguments for ttk.Labelframe
        """
        #: str: The title of the frame
        self.title = "Channel Settings"
        ttk.Labelframe.__init__(self, channels_tab, text=self.title, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: list: List of the variables for the channel checkbuttons
        self.channel_variables = []

        #: list: List of the channel checkbuttons
        self.channel_checks = []

        #: list: List of the variables for the laser dropdowns
        self.laser_variables = []

        #: list: List of the laser dropdowns
        self.laser_pulldowns = []

        #: list: List of the variables for the laser power dropdowns
        self.laserpower_variables = []

        #: list: List of the laser power dropdowns
        self.laserpower_pulldowns = []

        #: list: List of the variables for the filterwheel dropdowns
        self.filterwheel_variables = []

        #: list: List of the filterwheel dropdowns
        self.filterwheel_pulldowns = []

        #: list: List of the variables for the exposure time dropdowns
        self.exptime_variables = []

        #: list: List of the exposure time dropdowns
        self.exptime_pulldowns = []

        #: list: List of the variables for the time interval spinboxes
        self.interval_variables = []

        #: list: List of the time interval spinboxes
        self.interval_spins = []

        #: list: List of the variables for the defocus spinboxes
        self.defocus_variables = []

        #: list: List of the defocus spinboxes
        self.defocus_spins = []

        #: list: List of the labels for the columns
        self.label_text = [
            "Channel",
            "Laser",
            "Power",
            "Filter",
            "Exp. Time (ms)",
            "Interval",
            "Defocus",
        ]

        #: list: List of the labels for the columns
        self.labels = []

        #: list: List of the frames for the columns
        self.frame_columns = []

    def populate_frame(self, channels):
        """Populates the frame with the widgets.

        This function populates the frame with the widgets for the channels. By updating
        the self.label_text list, the columns can be changed. This function is called
        when the frame is initialized, and when the number of channels is changed in the
        controller.

        Parameters
        ----------
        channels : int
            The number of channels to be added to the frame.
        """

        #  Creates a column frame for each widget,
        for idx in range(len(self.label_text)):
            self.frame_columns.append(ttk.Frame(self))
            self.frame_columns[idx].columnconfigure(0, weight=1, uniform=1)
            self.frame_columns[idx].rowconfigure("all", weight=1, uniform=1)
            self.frame_columns[idx].grid(
                row=0, column=idx, sticky=tk.NSEW, padx=1, pady=(4, 6)
            )
            self.labels.append(
                ttk.Label(self.frame_columns[idx], text=self.label_text[idx])
            )
            self.labels[idx].grid(row=0, column=0, sticky=tk.N, pady=1, padx=1)
        self.frame_columns[5].grid(padx=(1, 4))
        self.frame_columns[0].grid(padx=(4, 1))

        # Creates the widgets for each channel - populates the rows.
        for num in range(0, channels):

            #  Channel Checkboxes
            self.channel_variables.append(tk.BooleanVar())
            self.channel_checks.append(
                ttk.Checkbutton(
                    self.frame_columns[0],
                    text="CH" + str(num + 1),
                    variable=self.channel_variables[num],
                )
            )
            self.channel_checks[num].grid(
                row=num + 1, column=0, sticky=tk.NSEW, padx=1, pady=1
            )

            #  Laser Dropdowns
            self.laser_variables.append(tk.StringVar())
            self.laser_pulldowns.append(
                ttk.Combobox(
                    self.frame_columns[1],
                    textvariable=self.laser_variables[num],
                    width=6,
                )
            )
            self.laser_pulldowns[num].config(state = "readonly")
            self.laser_pulldowns[num].grid(
                row=num + 1, column=0, sticky=tk.NSEW, padx=1, pady=1
            )

            #  Laser Power Spinbox
            self.laserpower_variables.append(tk.StringVar())
            self.laserpower_pulldowns.append(
                ValidatedSpinbox(
                    self.frame_columns[2],
                    from_=0,
                    to=100.0,
                    textvariable=self.laserpower_variables[num],
                    increment=5,
                    width=3,
                    font=tk.font.Font(size=11),
                )
            )
            self.laserpower_pulldowns[num].grid(
                row=num + 1, column=0, sticky=tk.NS, padx=1, pady=1
            )

            #  FilterWheel Dropdowns
            self.filterwheel_variables.append(tk.StringVar())
            self.filterwheel_pulldowns.append(
                ttk.Combobox(
                    self.frame_columns[3],
                    textvariable=self.filterwheel_variables[num],
                    width=10,
                )
            )
            self.filterwheel_pulldowns[num].config(state = "readonly")
            self.filterwheel_pulldowns[num].grid(
                row=num + 1, column=0, sticky=tk.NSEW, padx=1, pady=1
            )

            #  Exposure Time Spin boxes
            self.exptime_variables.append(tk.StringVar())
            self.exptime_pulldowns.append(
                ValidatedSpinbox(
                    self.frame_columns[4],
                    from_=0,
                    to=5000.0,
                    textvariable=self.exptime_variables[num],
                    increment=25,
                    width=5,
                    font=tk.font.Font(size=11),
                )
            )
            self.exptime_pulldowns[num].grid(
                row=num + 1, column=0, sticky=tk.NSEW, padx=1, pady=1
            )

            #  Time Interval Spin boxes
            self.interval_variables.append(tk.StringVar())
            self.interval_spins.append(
                ValidatedSpinbox(
                    self.frame_columns[5],
                    from_=0,
                    to=5000.0,
                    textvariable=self.interval_variables[num],
                    increment=1,
                    width=3,
                    font=tk.font.Font(size=11),
                )
            )
            self.interval_spins[num].grid(
                row=num + 1, column=0, sticky=tk.NSEW, padx=1, pady=1
            )
            hover = Hover(self.interval_spins[num], text="Not Implemented", type="free")
            hover.setdescription(text="Not Implemented")

            # Defocus Spinbox
            self.defocus_variables.append(tk.DoubleVar())
            self.defocus_spins.append(
                ValidatedSpinbox(
                    self.frame_columns[6],
                    from_=-500.0,
                    to=500.0,
                    textvariable=self.defocus_variables[num],
                    increment=0.1,
                    width=4,
                    font=tk.font.Font(size=11),
                )
            )
            self.defocus_spins[num].grid(
                row=num + 1, column=0, sticky=tk.NSEW, padx=1, pady=1
            )

            if num % 2 == 1:
                self.filterwheel_pulldowns[num].grid(pady=2)
                self.laser_pulldowns[num].grid(pady=2)
                self.channel_checks[num].grid(pady=2)


class StackAcquisitionFrame(ttk.Labelframe):
    """This class is the frame that holds the stack acquisition settings."""

    def __init__(self, settings_tab, *args, **kwargs):
        """Initialization of the Stack Acquisition Frame

        Parameters
        ----------
        settings_tab : tkinter.Frame
            The frame that holds the settings tab.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Frame
        text_label = "Stack Acquisition Settings (" + "\N{GREEK SMALL LETTER MU}" + "m)"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: dict: Dictionary of the widgets in the frame
        self.inputs = {}

        #: dict: Dictionary of the buttons in the frame
        self.buttons = {}

        #: tkinter.Frame: The frame that holds the position and slice settings
        self.pos_slice = ttk.Frame(self)

        #: tkinter.Frame: The frame that holds the laser cycling settings
        self.cycling = ttk.Frame(self)

        # Griding Each Holder Frame
        self.pos_slice.grid(row=0, column=0, sticky=tk.NSEW)
        self.cycling.grid(row=1, column=0, sticky=tk.NSEW)

        # Start Pos Frame (Vertically oriented)
        start_names = ["start_position", "start_focus"]
        start_labels = ["Pos", "Foc"]

        #: tkinter.Label: The label for the start position frame
        self.start_label = ttk.Label(self.pos_slice, text="Start")
        self.start_label.grid(row=0, column=0, sticky="S")
        for i in range(len(start_names)):
            self.inputs[start_names[i]] = LabelInput(
                parent=self.pos_slice,
                label=start_labels[i],
                input_class=ValidatedSpinbox,
                input_var=tk.DoubleVar(),
                input_args={"from_": 0.0, "to": 10000, "increment": 0.5, "width": 6},
            )
            self.inputs[start_names[i]].grid(
                row=i + 1, column=0, sticky="N", pady=2, padx=(6, 0)
            )
            self.inputs[start_names[i]].label.grid(sticky="N")

        # Start button
        self.buttons["set_start"] = HoverButton(
            self.pos_slice, text="Set Start Pos/Foc"
        )
        self.buttons["set_start"].grid(row=3, column=0, sticky="N", pady=2, padx=(6, 0))

        # End Pos Frame (Vertically Oriented)
        end_names = ["end_position", "end_focus"]
        end_labels = ["Pos", "Foc"]

        #: tkinter.Label: The label for the end position
        self.end_label = ttk.Label(self.pos_slice, text="End")
        self.end_label.grid(row=0, column=1, sticky="S")
        for i in range(len(end_names)):
            self.inputs[end_names[i]] = LabelInput(
                parent=self.pos_slice,
                label=end_labels[i],
                input_class=ValidatedSpinbox,
                input_var=tk.DoubleVar(),
                input_args={"from_": 0.0, "to": 10000, "increment": 0.5, "width": 6},
            )
            self.inputs[end_names[i]].grid(
                row=i + 1, column=1, sticky="N", pady=2, padx=(6, 0)
            )
            self.inputs[end_names[i]].label.grid(sticky="N")

        # End Button
        self.buttons["set_end"] = HoverButton(self.pos_slice, text="Set End Pos/Foc")
        self.buttons["set_end"].grid(row=3, column=1, sticky="N", pady=2, padx=(6, 0))

        # Step Size Frame (Vertically oriented)
        #: tkinter.Label: The label for the step size
        self.step_size_label = ttk.Label(self.pos_slice, text="Step Size")
        self.step_size_label.grid(row=0, column=2, sticky="S")
        self.inputs["step_size"] = LabelInput(
            parent=self.pos_slice,
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 6},
        )
        self.inputs["step_size"].grid(row=1, column=2, sticky="N", padx=6)

        # Slice Frame (Vertically oriented)
        #: tkinter.Label: The label to add empty space to the slice frame
        self.empty_label = ttk.Label(self.pos_slice, text=" ")
        self.empty_label.grid(row=0, column=3, sticky="N")
        slice_names = ["number_z_steps", "abs_z_start", "abs_z_end"]
        slice_labels = ["# slices      ", "Abs Z Start", "Abs Z Stop"]
        for i in range(len(slice_names)):
            self.inputs[slice_names[i]] = LabelInput(
                parent=self.pos_slice,
                label=slice_labels[i],
                input_class=ttk.Spinbox,
                input_var=tk.DoubleVar(),
                input_args={"increment": 0.5, "width": 6},
            )
            self.inputs[slice_names[i]].widget.configure(state="disabled")
            self.inputs[slice_names[i]].grid(
                row=i + 1, column=3, sticky="NSEW", pady=2, padx=(6, 0)
            )

        # Laser Cycling Settings
        self.inputs["cycling"] = LabelInput(
            parent=self.cycling,
            label="Laser Cycling Settings ",
            input_class=ValidatedCombobox,
            input_var=tk.StringVar(),
            input_args={"width": 8},
        )
        self.inputs["cycling"].state(
            ["readonly"]
        )  # Makes it so the user cannot type a choice into combobox
        self.inputs["cycling"].grid(row=0, column=0, sticky="NSEW", padx=6, pady=5)

        # Initialize DescriptionHovers
        self.inputs["step_size"].widget.hover.setdescription("Step Size")
        self.buttons["set_end"].hover.setdescription("Sets End")

    # Getters
    def get_variables(self):
        """Returns a dictionary of the variables in the widget

        This function returns a dictionary of all the variables
        that are tied to each widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of the variables in the widget
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Returns a dictionary of the widgets.

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary of the widgets in the widget
        """
        return self.inputs

    def get_buttons(self):
        """Returns a dictionary of the buttons in the frame.

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary of the buttons in the widget
        """
        return self.buttons


class StackTimePointFrame(ttk.Labelframe):
    """Frame for the stack timepoint settings in the channels tab.

    This class is a frame that holds the widgets for the stack timepoint settings.
    It is a subclass of ttk.Labelframe.
    """

    def __init__(self, settings_tab, *args, **kwargs):
        """Initilization of the Stack Timepoint Frame

        Parameters
        ----------
        settings_tab : tk.Frame
            The frame that the stack timepoint frame will be placed in
        *args : tuple
            Variable length argument list
        **kwargs : dict
            Arbitrary keyword arguments
        """
        text_label = "Timepoint Settings"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Widget dictionary
        #: dict: Dictionary of the widgets in the frame
        self.inputs = {}

        # Save Data Label
        #: tkinter.Label: The label for the save data checkbox
        self.laser_label = ttk.Label(self, text="Save Data")
        self.laser_label.grid(row=0, column=0, sticky=tk.NSEW, padx=(4, 5), pady=(4, 0))

        # Save Data Checkbox
        #: tk.BooleanVar: The variable for the save data checkbox
        self.save_data = tk.BooleanVar()
        self.save_data.set(False)
        #: ttk.Checkbutton: The save data checkbox
        self.save_check = ttk.Checkbutton(self, text="", variable=self.save_data)
        self.save_check.grid(row=0, column=1, sticky=tk.NSEW, pady=(4, 0))
        self.inputs["save_check"] = self.save_check

        # Timepoints Label, spinbox defaults to 1.
        #: tkinter.Label: The label for the timepoints spinbox
        self.filterwheel_label = ttk.Label(self, text="Timepoints")
        self.filterwheel_label.grid(
            row=1, column=0, sticky=tk.NSEW, padx=(4, 5), pady=2
        )
        #: tk.StringVar: The variable for the timepoints spinbox
        self.exp_time_spinval = tk.StringVar()
        #: ttk.Spinbox: The timepoints spinbox
        self.exp_time_spinbox = ValidatedSpinbox(
            self,
            from_=0,
            to=5000,
            textvariable=self.exp_time_spinval,
            increment=1,
            width=3,
        )
        self.exp_time_spinbox.grid(row=1, column=1, sticky=tk.NSEW, pady=2)
        self.inputs["time_spin"] = self.exp_time_spinbox

        # Stack Acq. Time Label
        #: tkinter.Label: The label for the stack acquisition time spinbox
        self.exp_time_label = ttk.Label(self, text="Stack Acq. Time")
        self.exp_time_label.grid(row=2, column=0, sticky=tk.NSEW, padx=(4, 5), pady=2)

        # Stack Acq. Time Spinbox
        #: tk.StringVar: The variable for the stack acquisition time spinbox
        self.stack_acq_spinval = tk.StringVar()
        #: ttk.Spinbox: The stack acquisition time spinbox
        self.stack_acq_spinbox = ttk.Spinbox(
            self,
            from_=0,
            to=5000.0,
            textvariable=self.stack_acq_spinval,  # this holds the data in the entry
            increment=25,
            width=6,
        )
        self.stack_acq_spinbox.grid(row=2, column=1, sticky=tk.NSEW, pady=2)
        self.stack_acq_spinbox.state(["disabled"])  # Starts it disabled

        # Stack Pause Label
        #: tkinter.Label: The label for the stack pause spinbox
        self.exp_time_label = ttk.Label(self, text="Stack Pause (s)")
        self.exp_time_label.grid(row=0, column=2, sticky=tk.NSEW, padx=(4, 5), pady=2)

        # Stack Pause Spinbox
        #: tk.StringVar: The variable for the stack pause spinbox
        self.stack_pause_spinval = tk.StringVar()
        #: ttk.Spinbox: The stack pause spinbox
        self.stack_pause_spinbox = ValidatedSpinbox(
            self,
            from_=0,
            to=5000.0,
            textvariable=self.stack_pause_spinval,
            increment=25,
            width=6,
        )
        self.stack_pause_spinbox.grid(row=0, column=3, sticky=tk.NSEW, pady=2)
        self.inputs["stack_pause"] = self.stack_pause_spinbox

        # Timepoint Interval Label
        #: tkinter.Label: The label for the timepoint interval spinbox
        self.exp_time_label = ttk.Label(self, text="Time Interval (hh:mm:ss)")
        self.exp_time_label.grid(row=1, column=2, sticky=tk.NSEW, padx=(4, 5), pady=2)

        # Timepoint Interval Spinbox
        #: tk.StringVar: The variable for the timepoint interval spinbox
        self.timepoint_interval_spinval = tk.StringVar()
        if self.timepoint_interval_spinval.get() == "":
            self.timepoint_interval_spinval.set("0")
        #: ttk.Spinbox: The timepoint interval spinbox
        self.timepoint_interval_spinbox = ttk.Spinbox(
            self,
            from_=0,
            to=5000.0,
            textvariable=self.timepoint_interval_spinval,
            increment=25,
            width=6,
        )
        self.timepoint_interval_spinbox.grid(row=1, column=3, sticky=tk.NSEW, pady=2)
        self.timepoint_interval_spinbox.state(["disabled"])  # Starts it disabled

        # Total Time Label
        #: tkinter.Label: The label for the total time spinbox
        self.exp_time_label = ttk.Label(self, text="Experiment Duration (hh:mm:ss)")
        self.exp_time_label.grid(
            row=2, column=2, sticky=tk.NSEW, padx=(4, 5), pady=(2, 6)
        )

        # Total Time Spinbox
        #: tk.StringVar: The variable for the total time spinbox
        self.total_time_spinval = tk.StringVar()
        if self.total_time_spinval.get() == "":
            self.total_time_spinval.set("0")
        #: ttk.Spinbox: The total time spinbox
        self.total_time_spinval = ttk.Spinbox(
            self,
            from_=0,
            to=5000.0,
            textvariable=self.total_time_spinval,
            increment=25,
            width=6,
        )
        self.total_time_spinval.grid(row=2, column=3, sticky=tk.NSEW, pady=(2, 6))
        self.total_time_spinval.state(["disabled"])

    # Getters
    def get_variables(self):
        """Returns a dictionary of all the variables that are tied to each widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            A dictionary of all the variables that are tied to each widget name.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Returns a dictionary of all the widgets that are tied to each widget name.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        widgets : dict
            A dictionary of all the widgets that are tied to each widget name.
        """
        return self.inputs


class MultiPointFrame(ttk.Labelframe):
    """Multi-Position Acquisition Frame"""

    def __init__(self, settings_tab, *args, **kwargs):
        """Initilization of the Multi-Position Acquisition Frame

        Parameters
        ----------
        settings_tab : tk.Frame
            The frame that the multipoint frame will be placed in
        *args : tuple
            Variable length argument list
        **kwargs : dict
            Arbitrary keyword arguments
        """
        text_label = "Multi-Position Acquisition"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Save Data Label
        #: tkinter.Label: The label for the save data checkbox
        self.laser_label = ttk.Label(self, text="Enable")
        self.laser_label.grid(row=0, column=0, sticky=tk.NSEW, padx=(4, 4), pady=(4, 4))

        # Save Data Checkbox
        #: tk.BooleanVar: The variable for the save data checkbox
        self.on_off = tk.BooleanVar()
        #: ttk.Checkbutton: The save data checkbox
        self.save_check = ttk.Checkbutton(self, text="", variable=self.on_off)
        self.save_check.grid(row=0, column=1, sticky=tk.NSEW, pady=(4, 4))

        # Tiling Wizard Button
        #: dict: Dictionary of the buttons in the frame
        self.buttons = {}
        self.buttons["tiling"] = ttk.Button(self, text="Launch Tiling Wizard")
        self.buttons["tiling"].grid(
            row=0, column=2, sticky=tk.NSEW, padx=(10, 0), pady=(4, 4)
        )

    def get_variables(self):
        """Returns a dictionary of all the variables that are tied to each widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            A dictionary of all the variables that are tied to each widget name.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Returns a dictionary of all the widgets that are tied to each widget name.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        widgets : dict
            A dictionary of all the widgets that are tied to each widget name.
        """
        return self.inputs


class QuickLaunchFrame(ttk.Labelframe):
    """Quick Launch Buttons Frame

    This frame contains buttons that launch the Tiling Wizard.
    """

    def __init__(self, settings_tab, *args, **kwargs):
        """Initilization of the Quick Launch Buttons Frame

        Parameters
        ----------
        settings_tab : object
            The settings tab object that this frame is being added to.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        text_label = "Quick Launch Buttons"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Tiling Wizard Button
        #: dict: Dictionary of the buttons in the frame
        self.buttons = {
            "waveform_parameters": ttk.Button(self, text="Waveform Parameters")
        }
        self.buttons["waveform_parameters"].grid(
            row=0, column=2, sticky=tk.NSEW, padx=(4, 4), pady=(4, 4)
        )

        self.buttons["autofocus_button"] = ttk.Button(self, text="Autofocus Settings")
        self.buttons["autofocus_button"].grid(
            row=1, column=2, sticky=tk.NSEW, padx=(4, 4), pady=(4, 4)
        )



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import tkinter as tk
from tkinter import ttk

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AcquireBar(ttk.Frame):
    """Acquire Bar Class.

    Class for the acquisition bar found at the top of the main application window.
    Main function is to change acq setting and then call the acquisition top-
    level window
    """

    def __init__(self, top_frame, root, *args, **kwargs):
        """Initialize Acquire Bar.

        Parameters
        ----------
        top_frame : tk.Frame
            Frame to place the acquire bar in.
        root : tk.Tk
            Root window of the application.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """
        #  Init bar with frame attr
        ttk.Frame.__init__(self, top_frame, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Putting bar into frame
        self.grid(row=0, column=0)

        # Acquire Button
        #: ttk.Button: Button to start acquisition
        self.acquire_btn = ttk.Button(self, text="Acquire")

        # Read Only Pull down menu: continuous, z-stack, single acquisition, projection.
        #: tk.StringVar: Variable to hold the current option selected
        self.options = tk.StringVar()
        #: ttk.Combobox: Pull down menu to select acquisition type
        self.pull_down = ttk.Combobox(self, textvariable=self.options)
        self.pull_down.state(["readonly"])

        #  Progress Bar: Current Acquisitions and Overall
        #: ttk.Frame: Frame to hold the progress bars
        self.progBar_frame = ttk.Frame(self)

        #  This is used to hold and grid the two progress bars.Now when this is
        #  loaded into acquisition bar, the progress bars will follow
        #: ttk.Progressbar: Progress bar for current acquisition
        self.CurAcq = ttk.Progressbar(
            self.progBar_frame, orient=tk.HORIZONTAL, length=200, mode="indeterminate"
        )

        #  Change mode to determinate and set steps for more intuitive usage
        #: ttk.Progressbar: Progress bar for overall acquisition
        self.OvrAcq = ttk.Progressbar(
            self.progBar_frame, orient=tk.HORIZONTAL, length=200, mode="determinate"
        )

        #: tk.Label: Label to display the current acquisition progress
        self.total_acquisition_label = tk.Label(self, text=f"{0:02}"
                                                           f":{0:02}"
                                                           f":{0:02}")

        self.CurAcq.grid(row=0, column=0)
        self.OvrAcq.grid(row=1, column=0)
        self.total_acquisition_label.grid(row=0, column=3, sticky=tk.NSEW)

        #: ttk.Button: Button to exit the application
        self.exit_btn = ttk.Button(self, text="Exit")

        #: ttk.Button: Button to stop the stage
        self.stop_stage = ttk.Button(self, text="Stop Stage")

        self.acquire_btn.grid(row=0, column=0, sticky=tk.NSEW, pady=(2, 2), padx=(2, 2))
        self.pull_down.grid(row=0, column=1, sticky=tk.NSEW, pady=(2, 2), padx=(2, 2))
        self.progBar_frame.grid(
            row=0, column=2, sticky=tk.NSEW, pady=(2, 2), padx=(2, 2)
        )
        self.stop_stage.grid(row=0, column=4, sticky=tk.NSEW, pady=(2, 2), padx=(2, 2))
        self.exit_btn.grid(row=0, column=5, sticky=tk.NSEW, pady=(2, 2), padx=(2, 2))
